/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include <wtf/Platform.h>

#include <PlatformOS.h>
using namespace LibPlatform;

#include "EventHandler.h"
#include "Frame.h"
#include "Node.h"
#include "RenderView.h"
#include "RenderLayer.h"

#include "WiiUAutoScroller.h"
#include "WiiUScrollConstants.h"

WiiUAutoScroller::WiiUAutoScroller(WebCore::Node* node)
	:m_scrollNode(node),
	 m_XVelocity(0.0f),
	 m_YVelocity(0.0f),
	 m_lastTime(0.0f),
	 m_duration(0.0f),
	 m_acceleration(WiiUScrollConstants::default_acceleration()),
	 m_currentStepValue(0.0f),
	 m_startStepValue(0.0f),
	 m_endStepValue(0.0f),
	 m_animationDuration(0.0f),
	 m_animationInterval(0.0f),
	 m_elapsedTime(0.0f),
	 m_timer(this, &WiiUAutoScroller::scrollStep)
{

}

WiiUAutoScroller::~WiiUAutoScroller()
{
	stop();
}

void WiiUAutoScroller::start(float x_velocity, float y_velocity)
{
	if (m_acceleration >= 0.0f) {
		m_acceleration = WiiUScrollConstants::default_acceleration();
	}
	float v = std::max(fabs(x_velocity), fabs(y_velocity));
	m_lastTime = 0.0f;
	m_XVelocity = x_velocity;
	m_YVelocity = y_velocity;
	m_duration = (-v / m_acceleration);

	//start the timer
	m_startStepValue = m_currentStepValue;
	m_endStepValue = 1.0;

	if (m_duration == 0) {
		//skip to end of animation
	}else if (m_currentStepValue == m_endStepValue) {
		return;
	}

	m_animationDuration = m_duration * (1 - m_currentStepValue);
	m_animationInterval = calculateInterval(defaultFrameRateHz);
	m_timer.stop();

	m_isScrolling = true;
	m_timer.startRepeating(m_animationInterval);
}

void WiiUAutoScroller::stop()
{
	m_XVelocity = m_YVelocity = m_lastTime = m_duration = 0.0f;
	//stop the timer
	if (isScrolling()) {
		m_timer.stop();
		m_isScrolling = false;
	}
}

void WiiUAutoScroller::scrollEnded()
{
	stop();
}

void WiiUAutoScroller::scrollProgressed()
{
	float t = m_currentStepValue * m_duration;
	float a_x = m_XVelocity > 0 ? m_acceleration : -m_acceleration;
	float a_y = m_YVelocity > 0 ? m_acceleration : -m_acceleration;
	float dx = getDelta(m_XVelocity, a_x, m_lastTime, t);
	float dy = getDelta(m_YVelocity, a_y, m_lastTime, t);
	m_lastTime = t;
	scrollNode(m_scrollNode, -1.0f * dx, -1.0f * dy);
}

void WiiUAutoScroller::scrollCanceled()
{
	stop();
}

void WiiUAutoScroller::scrollStep(WebCore::Timer<WiiUAutoScroller>*)
{
	m_elapsedTime += m_animationInterval;
	m_currentStepValue = m_elapsedTime / m_animationDuration;
	if (m_currentStepValue >= 1.0) {
		m_currentStepValue = 1.0;
	}
	scrollProgressed();
	if (m_currentStepValue == 1.0) {
		stop();
	}
}

float WiiUAutoScroller::calculateInterval(int frameRate)
{
	float timerInterval = 1000000 / frameRate;
	if (timerInterval < 10000) {
		timerInterval = 10000;
	}
	return (timerInterval/1000000);
}

float WiiUAutoScroller::calculateTween(float state)
{
	//We're assuming animation to be ease out
	return 1.0f - pow(1.0f - state, 2);
}

float WiiUAutoScroller::getDelta(float v0, float a, float t1, float t2)
{
	return getPosition(v0, a, t2) - getPosition(v0, a, t1);
}

float WiiUAutoScroller::getPosition(float v0, float a, float t)
{
	float max_t = -v0/a;
	if (t > max_t) {
		t = max_t;
	}
	return t*(v0+0.5f*a*t);
}

void WiiUAutoScroller::scrollNode(WebCore::Node* node, int dx, int dy) {
	if (!node) {
		return;
	}

	WebCore::RenderObject* renderer = node->renderer();
	WebCore::RenderLayer* layer = renderer ? renderer->enclosingLayer() : 0;

	if (!layer) {
		return;
	}

	bool isLineClampNoneByParent = true;
	if (renderer->parent()) {
		isLineClampNoneByParent = renderer->parent()->style()->lineClamp().isNone();
	}
	if (renderer->hasOverflowClip() && isLineClampNoneByParent) {
		int offsetX = layer->scrollXOffset() + dx;
		int offsetY = layer->scrollYOffset() + dy;
		layer->scrollToOffset(offsetX, offsetY, WebCore::RenderLayer::ScrollOffsetClamped);

		WebCore::Frame* frame = renderer->frame();
		if(frame) {
			WebCore::EventHandler* eventHandler = frame->eventHandler();
			//eventHandler->cancelFakeMouseMoveEvent();
			eventHandler->updateAutoscrollRenderer();
		}
	}else if (renderer->view()->frameView()) {
		renderer->view()->frameView()->scrollBy(WebCore::IntSize(dx,dy));
	}
}

