/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "JSSystemEventDispatcher.h"
#include <wtf/Platform.h> // PLATFORM(...)

#include "wk_appinterface.h"
#include "WebViewWiiU.h"
#include "JSExtensionData.h"

#include <algorithm>
#include <string>
#ifdef _WIIU_
#include <shim.h>
#endif
#include <PlatformAccounts.h>
#include <PlatformError.h>
#include <PlatformInput.h>
#include <PlatformDefines.h>
#include <PlatformSCI.h>
#include <PlatformConfig.h>

using namespace LibPlatform;

#define ENUM_TO_STRING(e) # e
#define CHAR_0_VALUE 48

#define REMOTE_URCC_BUTTON_MASK_NO_HOME (REMOTE_URCC_BUTTON_MASK - REMOTE_BUTTON_HOME)
#define GAMEPAD_BUTTON_MASK_NO_HOME (GAMEPAD_BUTTON_MASK - GAMEPAD_BUTTON_HOME - GAMEPAD_BUTTON_HOME)
#define REMOTE_CL_URCC_BUTTON_MASK_NO_HOME (REMOTE_URCC_BUTTON_MASK - REMOTE_CL_BUTTON_HOME)

static std::vector<bool> *wii_remote_connected;
static std::vector<bool> *wii_remote_disconnected;
static std::vector<bool> *wii_remote_extension_connected;
static std::vector<bool> *wii_remote_extension_disconnected;
static std::vector<bool> *wii_remote_extension_lstick_was_active;
static std::vector<bool> *wii_remote_extension_rstick_was_active;
static std::vector<unsigned> *wii_remote_battery_levels;
static unsigned bboard_battery_level;
static std::vector<bool> *wii_remote_mpls_connected;
static std::vector<bool> *wii_remote_mpls_disconnected;
static std::vector<bool> *display_remote_connected;
static std::vector<bool> *display_remote_disconnected;
static std::vector<bool> *gamepad_lstick_was_active;
static std::vector<bool> *gamepad_rstick_was_active;
static std::vector<unsigned> *gamepad_battery_levels;
static std::vector<bool> *pro_controller_connected;
static std::vector<bool> *pro_controller_disconnected;
static std::vector<bool> *pro_controller_lstick_was_active;
static std::vector<bool> *pro_controller_rstick_was_active;
static std::vector<unsigned> *pro_controller_battery_levels;
static std::vector<WebViewEventState_t> *webview_event_state;

bool g_DestructingJSSystemEventDispatcher = false;

namespace {
	unsigned buffer = 0;
	unsigned remote_reads_behind = 0;
	unsigned gamepad_reads_behind = 0;

	bool parentalControlRestrictionDispatched = false;

	struct DelayedDispatch {
		std::string event_name;
		std::string event_string;
		void *data;
		WiiUSystemChannel indexID;
		WiiUSystemDevice deviceID;
		WiiUSystemControl controlID; 
	};

	std::vector<DelayedDispatch> *delayed_dispatches;

	bool wii_bboard_connected = true;
	bool wii_bboard_disconnected = true;
}

void ResetJSDispatchEventGlobals(void)
{
	delete wii_remote_connected;					wii_remote_connected = 0;
	delete wii_remote_disconnected;					wii_remote_disconnected = 0;
	delete wii_remote_extension_connected;			wii_remote_extension_connected = 0;
	delete wii_remote_extension_disconnected;		wii_remote_extension_disconnected = 0;
	delete wii_remote_extension_lstick_was_active;	wii_remote_extension_lstick_was_active = 0;
	delete wii_remote_extension_rstick_was_active;	wii_remote_extension_rstick_was_active = 0;
	delete wii_remote_battery_levels;				wii_remote_battery_levels = 0;
	delete pro_controller_connected;				pro_controller_connected = 0;
	delete pro_controller_disconnected;				pro_controller_disconnected = 0;
	delete pro_controller_battery_levels;			pro_controller_battery_levels = 0;
	delete pro_controller_lstick_was_active;		pro_controller_lstick_was_active = 0;
	delete pro_controller_rstick_was_active;		pro_controller_rstick_was_active = 0;

	bboard_battery_level = 0;
	delete wii_remote_mpls_connected;				wii_remote_mpls_connected = 0;
	delete wii_remote_mpls_disconnected;			wii_remote_mpls_disconnected = 0;
	delete display_remote_connected;				display_remote_connected = 0;
	delete display_remote_disconnected;				display_remote_disconnected = 0;
	delete gamepad_battery_levels;					gamepad_battery_levels = 0;
	delete gamepad_lstick_was_active;				gamepad_lstick_was_active = 0;
	delete gamepad_rstick_was_active;				gamepad_rstick_was_active = 0;
	delete webview_event_state;						webview_event_state = 0;
	::buffer = 0;
	::remote_reads_behind = 0;
	::gamepad_reads_behind = 0;
	::parentalControlRestrictionDispatched = false;
	
	delete ::delayed_dispatches;
	::delayed_dispatches = 0;
	::wii_bboard_connected = true;
	::wii_bboard_disconnected = true;
}

void InitJSDispatchEventGlobals(void)
{
	wii_remote_connected = new std::vector<bool>;
	wii_remote_disconnected = new std::vector<bool>;
	wii_remote_extension_connected = new std::vector<bool>;
	wii_remote_extension_disconnected = new std::vector<bool>;
	wii_remote_extension_lstick_was_active = new std::vector<bool>;
	wii_remote_extension_rstick_was_active = new std::vector<bool>;
	wii_remote_battery_levels = new std::vector<unsigned>;
	pro_controller_connected = new std::vector<bool>;
	pro_controller_disconnected = new std::vector<bool>;
	pro_controller_lstick_was_active = new std::vector<bool>;
	pro_controller_rstick_was_active = new std::vector<bool>;
	pro_controller_battery_levels = new std::vector<unsigned>;
	wii_remote_mpls_connected = new std::vector<bool>;
	wii_remote_mpls_disconnected = new std::vector<bool>;
	display_remote_connected = new std::vector<bool>;
	display_remote_disconnected = new std::vector<bool>;
	gamepad_lstick_was_active = new std::vector<bool>;
	gamepad_rstick_was_active = new std::vector<bool>;
	gamepad_battery_levels = new std::vector<unsigned>;
	webview_event_state = new std::vector<WebViewEventState_t>;
	::buffer = 0;
	::remote_reads_behind = 0;
	::gamepad_reads_behind = 0;
	::delayed_dispatches = new std::vector<DelayedDispatch>;
}

JSExtensionEvent::JSExtensionEvent(WiiUEventType_t event_id, std::string event_type, void *data, std::string event_string, 
		WiiUSystemDevice deviceID, WiiUSystemControl controlID, WiiUSystemChannel indexID)
		: JSEvent(event_id, event_type, data, event_string),
		indexID(indexID),
		deviceID(deviceID),
		controlID(controlID)
{
}

WiiUSystemChannel JSExtensionEvent::GetIndexID()
{
	return indexID;
}

WiiUSystemDevice JSExtensionEvent::GetDeviceID()
{
	return deviceID;
}

WiiUSystemControl JSExtensionEvent::GetControlID()
{
	return controlID;
}

void DispatchFunctorEvent(std::string event_name, std::string event_string, void *data) {
	if (g_DestructingJSSystemEventDispatcher) return;
	JSEvent js_event(WIIU_EMPTY_EVENT, event_name, (void *)data); 
	JSSystemEventDispatcher::GetInstance()->DispatchEvent(js_event, event_string);
}

void QueueDispatchFunctorEvent(std::string event_name, std::string event_string, void *data, 
	WiiUSystemChannel indexID = CHANNEL_0, WiiUSystemDevice deviceID = REMOTE, WiiUSystemControl controlID = DEFAULT) {
	
	if (g_DestructingJSSystemEventDispatcher) return;

	DelayedDispatch dd = {event_name, event_string, data, indexID, deviceID, controlID};
	delayed_dispatches->push_back(dd);
}

void DispatchRemoteTrigger(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasRemoteTriggered(channel, ::buffer);

		if(trigger_data < 0xffff && trigger_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = trigger_data;

			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= trigger_data)
			{
				if(buttonBitFlag & trigger_data)
				{
					key_data.key_state = buttonBitFlag;

					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"press",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"press"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchRemoteRelease(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned release_data = input->WasRemoteReleased(channel, ::buffer);

		if(release_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = release_data;
			
			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= release_data)
			{
				if(buttonBitFlag & release_data)
				{
					key_data.key_state = buttonBitFlag;
					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"release",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"release"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchRemoteTrigOrRel(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {

		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasRemoteTriggered(channel, ::buffer);
		unsigned release_data = input->WasRemoteReleased(channel, ::buffer);

		// HACK: Fix it!
		if(trigger_data > 0xffff || release_data > 0xffff) return;

		if(trigger_data < 0xffff && (trigger_data | release_data) & (REMOTE_BUTTON_LEFT | REMOTE_BUTTON_UP | REMOTE_BUTTON_RIGHT | REMOTE_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsRemotePressed(channel) & (REMOTE_BUTTON_LEFT | REMOTE_BUTTON_UP | REMOTE_BUTTON_RIGHT | REMOTE_BUTTON_DOWN);

			if(press_data == REMOTE_BUTTON_LEFT) 								{key_data.x = -1; key_data.y =  0;}
			else if(press_data == REMOTE_BUTTON_UP) 							{key_data.x =  0; key_data.y =  1;}
			else if(press_data == REMOTE_BUTTON_RIGHT) 							{key_data.x =  1; key_data.y =  0;}
			else if(press_data == REMOTE_BUTTON_DOWN) 							{key_data.x =  0; key_data.y = -1;}
			else if(press_data == (REMOTE_BUTTON_LEFT | REMOTE_BUTTON_UP)) 		{key_data.x = -1; key_data.y =  1;}
			else if(press_data == (REMOTE_BUTTON_UP | REMOTE_BUTTON_RIGHT)) 	{key_data.x =  1; key_data.y =  1;}
			else if(press_data == (REMOTE_BUTTON_RIGHT | REMOTE_BUTTON_DOWN)) 	{key_data.x =  1; key_data.y = -1;}
			else if(press_data == (REMOTE_BUTTON_DOWN | REMOTE_BUTTON_LEFT)) 	{key_data.x = -1; key_data.y = -1;}
			else 																{key_data.x =  0; key_data.y =  0;}

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DIRECTION
			);
		}
	}
}

void DispatchRemoteAcceleration(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 dir_data = input->GetRemoteAccelerationDirection(channel, ::buffer);
		float mag_data = input->GetRemoteAccelerationMagnitude(channel, ::buffer);
		float change_data = input->GetRemoteAccelerationChange(channel, ::buffer);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data,
			(WiiUSystemChannel)channel,
			REMOTE,
			ACCELEROMETER
		);
	}
}

void DispatchRemoteGyro(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasRemoteController(channel)) {
		if(!input->IsRemoteEnabled(channel) || input->GetRemoteMode(channel) != LibPlatform::Input::WII_REMOTE_MODE_FULL || !input->RemoteHasMotionPlusInserted(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"update"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			REMOTE,
			GYRO
		);
	}
}

void DispatchRemoteGyroCalibration(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	//OS::Report("input->GetMotionPlusCalibrationStatus %f \n", input->GetMotionPlusCalibrationStatus(channel));

	if(::buffer < remote_reads_behind && input->GetMotionPlusCalibrationStatus(channel) == 0.0f) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"calibrationSuccess",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationSuccess"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			REMOTE,
			GYRO
		);
	}
}

void DispatchRemoteGyroCalibrationFailed(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	//OS::Report("input->GetMotionPlusCalibrationStatus %f \n", input->GetMotionPlusCalibrationStatus(channel));

	if(::buffer < remote_reads_behind && input->GetMotionPlusCalibrationFailed(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetRemoteAngularVelocity(channel, ::buffer);
		float3 angle_data = input->GetRemoteAngle(channel, ::buffer);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"calibrationFail",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationFail"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			REMOTE,
			GYRO
		);
	}
}

void DispatchRemotePointer(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer < remote_reads_behind && input->CheckRemotePointerPosition(channel, ::buffer)) {
		
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
	
		WebView *pView = (WebView*)WebViewTV();
		if (pView) {
			float2 position_data = input->GetRemotePointerPosition(channel, pView->width(), pView->height(), ::buffer);
			float distance_data = input->GetRemoteDistance(channel, ::buffer);
		
			position_data.x += pView->width() / 2.0f;
			position_data.y += pView->height() / 2.0f;

			//Get the pointer position from -1 to 1
			float2 raw_position_data = input->GetRemotePointerPosition(channel, 2, 2, ::buffer);

			WiiUMoveData move_data;
			move_data.channel = channel;
			move_data.screenX = position_data.x;
			move_data.screenY = position_data.y;
			move_data.x = raw_position_data.x;
			move_data.y = raw_position_data.y;
			move_data.z = distance_data;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"update",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DPD,
					"update"
				), 
				(void *)&move_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DPD
			);
		}
	}
}

void DispatchRemoteConnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(wii_remote_connected->size() <= channel) {
		wii_remote_connected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasRemoteController(channel)) {

		if(!(*wii_remote_connected)[channel]) {
			(*wii_remote_connected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	} else {
		(*wii_remote_connected)[channel] = false;
	}
}

void DispatchRemoteDisconnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(wii_remote_disconnected->size() <= channel) {
		wii_remote_disconnected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasRemoteController(channel)) {
		if(!(*wii_remote_disconnected)[channel]) {
			(*wii_remote_disconnected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerDisconnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	} else {
		(*wii_remote_disconnected)[channel] = false;
	}
}

void DispatchBBoardConnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasBalanceBoard()) {

		if(!::wii_bboard_connected) {
			::wii_bboard_connected = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					BALANCEBOARD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				BALANCEBOARD,
				DEFAULT
			);
		}
	} else {
		::wii_bboard_connected = false;
	}
}

void DispatchBBoardDisconnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasBalanceBoard()) {
		if(!wii_bboard_disconnected) {
			wii_bboard_disconnected = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					BALANCEBOARD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerDisconnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				BALANCEBOARD,
				DEFAULT
			);
		}
	} else {
		wii_bboard_disconnected = false;
	}
}

void DispatchRemoteExtensionAddedRemoved(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(wii_remote_extension_connected->size() <= channel) {
		wii_remote_extension_connected->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasClassicController(channel) || input->HasNunChuckController(channel)) {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}
		if(!(*wii_remote_extension_connected)[channel] && input->IsExtensionReady(channel)) {
			(*wii_remote_extension_connected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"extensionControllerAdded",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"extensionControllerAdded"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	} 
	else {
		if(!input->IsRemoteEnabled(channel))
		{
			return;
		}

		if((*wii_remote_extension_connected)[channel] && input->IsExtensionReady(channel)) {
			(*wii_remote_extension_connected)[channel] = false;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"extensionControllerRemoved",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"extensionControllerRemoved"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	}
}

void DispatchRemoteBatteryLevelChanged(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(!input->HasRemoteController(channel)) return;

	// Expand the array to however many remotes we have
	while(wii_remote_battery_levels->size() <= channel) {
		wii_remote_battery_levels->push_back(input->GetRemoteBatteryLevel(channel));
	}

	if(!input->IsRemoteEnabled(channel)) {
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->GetRemoteBatteryLevel(channel) != (*wii_remote_battery_levels)[channel]) {
		(*wii_remote_battery_levels)[channel] = input->GetRemoteBatteryLevel(channel);
		WiiUScalarData scalar_data;
		scalar_data.channel = channel;
		scalar_data.s = 0.0f;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"batteryLevelChange",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"batteryLevelChange"
			), 
			(void *)&scalar_data,
			(WiiUSystemChannel)channel,
			REMOTE,
			DEFAULT
		);
	}
}

void DispatchRemoteMplsAttached(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(wii_remote_mpls_connected->size() <= channel) {
		wii_remote_mpls_connected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(!input->IsRemoteEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->RemoteHasMotionPlusInserted(channel)) {
		if(!(*wii_remote_mpls_connected)[channel]) {
			(*wii_remote_mpls_connected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"mplsAdded",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"mplsAdded"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	} else {
		(*wii_remote_mpls_connected)[channel] = false;
	}
}

void DispatchRemoteMplsRemoved(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(wii_remote_mpls_disconnected->size() <= channel) {
		wii_remote_mpls_disconnected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(!input->IsRemoteEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(!input->RemoteHasMotionPlusInserted(channel)) {
		if(!(*wii_remote_mpls_disconnected)[channel]) {
			(*wii_remote_mpls_disconnected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"mplsRemoved",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"mplsRemoved"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				REMOTE,
				DEFAULT
			);
		}
	} else {
		(*wii_remote_mpls_disconnected)[channel] = false;
	}
}

void DispatchClassicTrigger(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasClassicTriggered(channel, ::buffer);
	
		if(trigger_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = trigger_data;
			
			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= trigger_data)
			{
				if(buttonBitFlag & trigger_data)
				{
					key_data.key_state = buttonBitFlag;
					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"press",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE_CL, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"press"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE_CL,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchClassicRelease(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		unsigned release_data = input->WasClassicReleased(channel, ::buffer);

		 if(release_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = release_data;
			
			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= release_data)
			{
				if(buttonBitFlag & release_data)
				{
					key_data.key_state = buttonBitFlag;
					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"release",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE_CL, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"release"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE_CL,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchClassicTrigOrRel(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->WasClassicTriggered(channel, ::buffer);
		unsigned release_data = input->WasClassicReleased(channel, ::buffer);

		// HACK: Fix it!
		if(trigger_data > 0xffff || release_data > 0xffff) return;

		if(trigger_data < 0xffff && (trigger_data | release_data) & (REMOTE_CL_BUTTON_LEFT | REMOTE_CL_BUTTON_UP | REMOTE_CL_BUTTON_RIGHT | REMOTE_CL_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsClassicPressed(channel) & (REMOTE_CL_BUTTON_LEFT | REMOTE_CL_BUTTON_UP | REMOTE_CL_BUTTON_RIGHT | REMOTE_CL_BUTTON_DOWN);

			if(press_data == REMOTE_CL_BUTTON_LEFT) 								{key_data.x = -1; key_data.y =  0;}
			else if(press_data == REMOTE_CL_BUTTON_UP) 								{key_data.x =  0; key_data.y =  1;}
			else if(press_data == REMOTE_CL_BUTTON_RIGHT) 							{key_data.x =  1; key_data.y =  0;}
			else if(press_data == REMOTE_CL_BUTTON_DOWN) 							{key_data.x =  0; key_data.y = -1;}
			else if(press_data == (REMOTE_CL_BUTTON_LEFT | REMOTE_CL_BUTTON_UP)) 	{key_data.x = -1; key_data.y =  1;}
			else if(press_data == (REMOTE_CL_BUTTON_UP | REMOTE_CL_BUTTON_RIGHT)) 	{key_data.x =  1; key_data.y =  1;}
			else if(press_data == (REMOTE_CL_BUTTON_RIGHT | REMOTE_CL_BUTTON_DOWN)) {key_data.x =  1; key_data.y = -1;}
			else if(press_data == (REMOTE_CL_BUTTON_DOWN | REMOTE_CL_BUTTON_LEFT)) 	{key_data.x = -1; key_data.y = -1;}
			else 																	{key_data.x =  0; key_data.y =  0;}

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data,
				(WiiUSystemChannel)channel,
				REMOTE_CL,
				DIRECTION
			);
		}
	}
}

void DispatchClassicLStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	while(wii_remote_extension_lstick_was_active->size() <= channel) {
		wii_remote_extension_lstick_was_active->push_back(false);
	}

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		float2 lstick_data = input->GetClassicLStick(channel, ::buffer);
		
		if(lstick_data.x || lstick_data.y || (*wii_remote_extension_lstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*wii_remote_extension_lstick_was_active)[channel] = lstick_data.x || lstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = lstick_data.x;
			vector2_data.y = lstick_data.y;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				REMOTE_CL,
				LSTICK
			);
		}
	}
}

void DispatchClassicRStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	while(wii_remote_extension_rstick_was_active->size() <= channel) {
		wii_remote_extension_rstick_was_active->push_back(false);
	}

	if(::buffer < remote_reads_behind && input->HasClassicController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		float2 rstick_data = input->GetClassicRStick(channel, ::buffer);

		if(rstick_data.x || rstick_data.y || (*wii_remote_extension_rstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*wii_remote_extension_rstick_was_active)[channel] = rstick_data.x || rstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = rstick_data.x;
			vector2_data.y = rstick_data.y;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_CL, 
					(WiiUSystemChannel)channel,
					RSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				REMOTE_CL,
				RSTICK
			);
		}
	}
}

void DispatchNunChuckPress(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}

		unsigned trigger_data = input->WasNunChuckTriggered(channel, ::buffer);

		if(trigger_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = trigger_data;

			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= trigger_data)
			{
				if(buttonBitFlag & trigger_data)
				{
					key_data.key_state = buttonBitFlag;
					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"press",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE_FS, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"press"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE_FS,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchNunChuckRelease(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		unsigned release_data = input->WasNunChuckReleased(channel, ::buffer);

		if(release_data) {
			WiiUKeyData key_data;
			key_data.channel = channel;
			key_data.key_state = release_data;

			unsigned buttonBitFlag = 1;
			unsigned buttonBitMax = 1 << 31;
			//This loop will fire events for each possible button
			while(buttonBitFlag <= release_data)
			{
				if(buttonBitFlag & release_data)
				{
					key_data.key_state = buttonBitFlag;
					JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
						"release",
						JSSystemEventDispatcher::GetInstance()->GetEventString(
							REMOTE_FS, 
							(WiiUSystemChannel)channel,
							BUTTON,
							"release"
						), 
						(void *)&key_data,
						(WiiUSystemChannel)channel,
						REMOTE_FS,
						BUTTON
					);
				}
				if(buttonBitFlag == buttonBitMax)
					break;
				buttonBitFlag = buttonBitFlag << 1;
			}
		}
	}
}

void DispatchNunChuckStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	while(wii_remote_extension_lstick_was_active->size() <= channel) {
		wii_remote_extension_lstick_was_active->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		float2 stick_data = input->GetNunChuckStick(channel, ::buffer);

		if(stick_data.x || stick_data.y || (*wii_remote_extension_lstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*wii_remote_extension_lstick_was_active)[channel] = stick_data.x || stick_data.y;

			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = stick_data.x;
			vector2_data.y = stick_data.y;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					REMOTE_FS, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				REMOTE_FS,
				LSTICK
			);
		}
	}
}

void DispatchNunChuckAccel(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasNunChuckController(channel)) {
		if(!input->IsExtensionEnabled(channel))
		{
			return;
		}
		float3 dir_data = input->GetNunChuckAccelerationDirection(channel, ::buffer);
		float mag_data = input->GetNunChuckAccelerationMagnitude(channel, ::buffer);
		float change_data = input->GetNunChuckAccelerationChange(channel, ::buffer);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				REMOTE_FS, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data,
			(WiiUSystemChannel)channel,
			REMOTE_FS,
			ACCELEROMETER
		);
	}
}

void DispatchGamePadTrigger(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->WasGamePadTriggered(channel, ::buffer);

		// HACK: Fix it!
		if(!trigger_data || trigger_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;


		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = trigger_data;

		unsigned buttonBitFlag = 1;
		unsigned buttonBitMax = 1 << 31;
		//This loop will fire events for each possible button
		while(buttonBitFlag <= trigger_data)
		{
			if(buttonBitFlag & trigger_data)
			{
				key_data.key_state = buttonBitFlag;
				JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
					"press",
					JSSystemEventDispatcher::GetInstance()->GetEventString(
						GAMEPAD, 
						(WiiUSystemChannel)channel,
						BUTTON,
						"press"
					), 
					(void *)&key_data,
					(WiiUSystemChannel)channel,
					GAMEPAD,
					BUTTON
				);
			}
			if(buttonBitFlag == buttonBitMax)
				break;
			buttonBitFlag = buttonBitFlag << 1;
		}
	}
}

void DispatchGamePadRelease(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned release_data = input->WasGamePadReleased(channel, ::buffer);

		// HACK: Fix it!
		if(!release_data || release_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;
		
		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = release_data;
			
		unsigned buttonBitFlag = 1;
		unsigned buttonBitMax = 1 << 31;
		//This loop will fire events for each possible button
		while(buttonBitFlag <= release_data)
		{
			if(buttonBitFlag & release_data)
			{
				key_data.key_state = buttonBitFlag;
				JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
					"release",
					JSSystemEventDispatcher::GetInstance()->GetEventString(
						GAMEPAD, 
						(WiiUSystemChannel)channel,
						BUTTON,
						"release"
					), 
					(void *)&key_data,
					(WiiUSystemChannel)channel,
					GAMEPAD,
					BUTTON
				);
			}
			if(buttonBitFlag == buttonBitMax)
				break;
			buttonBitFlag = buttonBitFlag << 1;
		}
		
	}
}


void DispatchGamePadTrigOrRel(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		unsigned trigger_data = input->IsGamePadTriggered(channel);
		unsigned release_data = input->IsGamePadReleased(channel);

		// HACK: Fix it!
		if(trigger_data > GAMEPAD_BUTTON_MASK_NO_HOME || release_data > GAMEPAD_BUTTON_MASK_NO_HOME) return;

		if((trigger_data | release_data) & (GAMEPAD_BUTTON_LEFT | GAMEPAD_BUTTON_UP | GAMEPAD_BUTTON_RIGHT | GAMEPAD_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->IsGamePadPressed(channel) & (GAMEPAD_BUTTON_LEFT | GAMEPAD_BUTTON_UP | GAMEPAD_BUTTON_RIGHT | GAMEPAD_BUTTON_DOWN);

			if(press_data == GAMEPADBASE_BUTTON_LEFT) 									{key_data.x = -1; key_data.y =  0;}
			else if(press_data == GAMEPADBASE_BUTTON_UP) 								{key_data.x =  0; key_data.y =  1;}
			else if(press_data == GAMEPADBASE_BUTTON_RIGHT) 							{key_data.x =  1; key_data.y =  0;}
			else if(press_data == GAMEPADBASE_BUTTON_DOWN) 								{key_data.x =  0; key_data.y = -1;}
			else if(press_data == (GAMEPADBASE_BUTTON_LEFT | GAMEPADBASE_BUTTON_UP)) 	{key_data.x = -1; key_data.y =  1;}
			else if(press_data == (GAMEPADBASE_BUTTON_UP | GAMEPADBASE_BUTTON_RIGHT)) 	{key_data.x =  1; key_data.y =  1;}
			else if(press_data == (GAMEPADBASE_BUTTON_RIGHT | GAMEPADBASE_BUTTON_DOWN)) {key_data.x =  1; key_data.y = -1;}
			else if(press_data == (GAMEPADBASE_BUTTON_DOWN | GAMEPADBASE_BUTTON_LEFT)) 	{key_data.x = -1; key_data.y = -1;}
			else 																		{key_data.x =  0; key_data.y =  0;}

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				DIRECTION
			);
		}
	}
}

void DispatchGamePadLStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;

	while(gamepad_lstick_was_active->size() <= channel) {
		gamepad_lstick_was_active->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 lstick_data = input->GetGamePadLStick(channel);

		if(lstick_data.x || lstick_data.y || (*gamepad_lstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*gamepad_lstick_was_active)[channel] = lstick_data.x || lstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = lstick_data.x;
			vector2_data.y = lstick_data.y;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				LSTICK
			);
		}
	}
}

void DispatchGamePadRStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;

	while(gamepad_rstick_was_active->size() <= channel) {
		gamepad_rstick_was_active->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 rstick_data = input->GetGamePadRStick(channel);

		if(rstick_data.x || rstick_data.y || (*gamepad_rstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*gamepad_rstick_was_active)[channel] = rstick_data.x || rstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = rstick_data.x;
			vector2_data.y = rstick_data.y;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					RSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				RSTICK
			);
		}
	}
}

void DispatchGamePadGyro(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float3 vel_data = input->GetGamePadGyro(channel);
		float3 angle_data = input->GetGamePadAngle(channel);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				GYRO,
				"update"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			GYRO
		);
	}
}

void DispatchGamePadGyroCalibration(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->GetGamePadGyroCalibrationSuccess(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetGamePadGyro(channel);
		float3 angle_data = input->GetGamePadAngle(channel);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"calibrationSuccess",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationSuccess"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			GYRO
		);
	}
}

void DispatchGamePadGyroCalibrationFailed(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->GetGamePadGyroCalibrationFailed(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}

		float3 vel_data = input->GetGamePadGyro(channel);
		float3 angle_data = input->GetGamePadAngle(channel);

		WiiUGyroData gyro_data;
		gyro_data.channel = channel;
		gyro_data.angle_x = angle_data.x;
		gyro_data.angle_y = angle_data.y;
		gyro_data.angle_z = angle_data.z;
		gyro_data.vel_x = vel_data.x;
		gyro_data.vel_y = vel_data.y;
		gyro_data.vel_z = vel_data.z;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"calibrationFail",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				GYRO,
				"calibrationFail"
			), 
			(void *)&gyro_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			GYRO
		);
	}
}

void DispatchGamePadPointer(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	//If Gamepad touch screen is being used, but this isn't the first frame of contact
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel) 
		&& input->CheckGamePadTouchPosition(channel) && !input->IsGamePadTouchTriggered(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 touch_position = input->GetGamePadTouchPosition(channel);
		WiiUVector2Data vector2_data;
		vector2_data.channel = channel;
		vector2_data.x = touch_position.x;
		vector2_data.y = touch_position.y;
		
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				TOUCH,
				"update"
			), 
			(void *)&vector2_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			TOUCH
		);
	}
}

void DispatchGamePadTouchStart(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer == gamepad_reads_behind && input->HasGamePadController(channel) && input->IsGamePadTouchTriggered(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 touch_position = input->GetGamePadTouchPosition(channel);
		WiiUVector2Data vector2_data;
		vector2_data.channel = channel;
		vector2_data.x = touch_position.x;
		vector2_data.y = touch_position.y;
		
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"touchStart",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				TOUCH,
				"touchStart"
			), 
			(void *)&vector2_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			TOUCH
		);
	}
}

void DispatchGamePadTouchEnd(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(::buffer == gamepad_reads_behind && input->HasGamePadController(channel) && input->IsGamePadTouchReleased(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float2 touch_position = input->GetGamePadTouchPosition(channel);
		WiiUVector2Data vector2_data;
		vector2_data.channel = channel;
		vector2_data.x = touch_position.x;
		vector2_data.y = touch_position.y;
		
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"touchEnd",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				TOUCH,
				"touchEnd"
			), 
			(void *)&vector2_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			TOUCH
		);
	}
}

void DispatchGamePadAcceleration(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < gamepad_reads_behind && input->HasGamePadController(channel)) {
		if(!input->IsGamePadEnabled(channel))
		{
			return;
		}
		float3 dir_data = input->GetGamePadAccelerationDirection(channel);
		float mag_data = input->GetGamePadAccelerationMagnitude(channel);
		float change_data = input->GetGamePadAccelerationChange(channel);

		WiiUAccelData accel_data;

		accel_data.channel = channel;
		accel_data.dir_x = dir_data.x;
		accel_data.dir_y = dir_data.y;
		accel_data.dir_z = dir_data.z;
		accel_data.mag = mag_data;
		accel_data.change = change_data;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"update",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				ACCELEROMETER,
				"update"
			), 
			(void *)&accel_data,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			ACCELEROMETER
		);
	}
}

void DispatchGamePadBatteryLevelChanged(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	// Expand the array to however many remotes we have
	while(gamepad_battery_levels->size() <= channel) {
		gamepad_battery_levels->push_back(input->GetGamePadBatteryLevel(channel));
	}

	if(!input->IsGamePadEnabled(channel))
	{
		return;
	}
	// If the state of the remote has changed send the appropriate event
	if(input->GetGamePadBatteryLevel(channel) != (*gamepad_battery_levels)[channel]) {
			(*gamepad_battery_levels)[channel] = input->GetGamePadBatteryLevel(channel);
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"batteryLevelChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"batteryLevelChange"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				DEFAULT
			);
	}
}


void DispatchGamePadConnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(display_remote_connected->size() <= channel) {
		display_remote_connected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasGamePadController(channel)) {
		if(!(*display_remote_connected)[channel]) {
			(*display_remote_connected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				DEFAULT
			);
		} 
	} else {
		(*display_remote_connected)[channel] = false;
	}
}

void DispatchGamePadDisconnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(display_remote_disconnected->size() <= channel) {
		display_remote_disconnected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasGamePadController(channel)) {
		if(!(*display_remote_disconnected)[channel]) {
			(*display_remote_disconnected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					GAMEPAD, 
					(WiiUSystemChannel)channel,
					DEFAULT, 
					"controllerDisconnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				GAMEPAD,
				DEFAULT
			);
		}
	} else {
		(*display_remote_disconnected)[channel] = false;
	}
}

void DispatchCAMReady(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel) && input->CameraReady(channel)) {
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"ready",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"ready"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			DEFAULT
		);
	}
}

void DispatchCAMStreamOpen(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel) && input->CameraStreamOpened(channel)) {
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"open",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"open"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			DEFAULT
		);
	}
}

void DispatchCAMStreamClose(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->CameraStreamClosed(channel)) {
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"close",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"close"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			DEFAULT
		);
	}
}

void DispatchCAMExit(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->CameraExited(channel)) {
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"exit",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				DEFAULT,
				"exit"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			DEFAULT
		);
	}
}

void DispatchMICReady(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicReadyEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"ready",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"ready"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}

void DispatchMICOpen(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicOpenEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"open",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"open"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}
void DispatchMICStreamUpdate(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasGamePadController(channel))
        input->MicUpdateGamePadMicrophone(channel);
}

void DispatchMICExit(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicExitEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"exit",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"exit"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}
void DispatchMICClose(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicCloseEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"close",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"close"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}
void DispatchMICConnectedUpdate(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicConnectedEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"connected",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"connected"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}
void DispatchMICSDisconnectedClose(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

    if(input->TestMicDisconnectedEvent(channel)){
        JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"disconnected",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				GAMEPAD,
				(WiiUSystemChannel)channel,
				MIC,
				"disconnected"
			), 
			0,
			(WiiUSystemChannel)channel,
			GAMEPAD,
			MIC
		);
    }
}


void DispatchErrors(signed) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Error *error = LibPlatform::Error::GetInstance();

	LibPlatform::Error::ReportInfo report_info;
	while(error->HasReports()) {
		error->NextReport(&report_info);

		WiiUErrorData error_data;
		error_data.code = report_info.code;
		std::swap(error_data.description, report_info.description);
		std::swap(error_data.originFunction, report_info.originFunction);
		error_data.originCode = report_info.originCode;
		// Extract the first three digits for the error type
		error_data.type = ((unsigned)report_info.type - ((unsigned)report_info.type % 10000)) / 10000;

		//APW - 12/4/12: Errors that have no code should have some usable value in the data field
		if(!report_info.code)
		{
			error_data.code = (unsigned)report_info.data;
		}

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"error",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				SYSTEM, 
				"error"
			), 
			(void *)&error_data,
			CHANNEL_0,
			SYSTEM
		);
	}
}

void DispatchCrash(signed) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// If we're crashing 
	if(JSSystemEventDispatcher::GetInstance()->GetCrash()) {
		// Let JS side know to save progress
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"crash",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				SYSTEM, 
				"crash"
			), 
			0,
			CHANNEL_0,
			SYSTEM
		);
	}
}

void DispatchBalanceBoardInfo(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasBalanceBoard()) {
		WiiUVector2Data weight_data;
		weight_data.channel = 3;
		weight_data.x = input->BalanceBoardGetRawWeight(channel);
		weight_data.y = input->BalanceBoardGetAvgWeight(channel);

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"weightChange",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				BALANCEBOARD, 
				(WiiUSystemChannel)channel,
				WEIGHT, 
				"weightChange"
			), 
			(void *)&weight_data,
			(WiiUSystemChannel)channel,
			BALANCEBOARD,
			WEIGHT
		);
	}
}

void DispatchCorrectedBalanceBoardInfo(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(input->HasBalanceBoard()) {
		WiiUScalarData weight_data;
		weight_data.channel = channel;
		weight_data.s = input->BalanceBoardGetTGCWeight();

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"weightChange",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				BALANCEBOARD, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"weightChange"
			), 
			(void *)&weight_data,
			(WiiUSystemChannel)channel,
			BALANCEBOARD,
			DEFAULT
		);
	}
}

void DispatchBalanceBoardBatteryLevelChanged(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(!input->HasBalanceBoard()) {
		return;
	}

	// If the state of the remote has changed send the appropriate event
	if(input->HasBalanceBoard() && input->BalanceBoardGetBatteryLevel() != bboard_battery_level) {
			bboard_battery_level = input->BalanceBoardGetBatteryLevel();
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = bboard_battery_level;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"batteryLevelChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					BALANCEBOARD, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"batteryLevelChange"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				BALANCEBOARD,
				DEFAULT
			);
	}
}

void DispatchParentalControlServiceApplicationRestricted(signed) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// If parental controls are restricted
	if(!::parentalControlRestrictionDispatched 
		&& LibPlatform::ApplicationConfig::GetIsServiceApplication()) {

		//This check may be computationally expensive, only check if other tests pass first
		if(LibPlatform::SCI::GetParentalControlServiceApplicationRestricted())
		{
			// Let JS side know to pop up warning
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"parentalControlServiceApplicationRestricted",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					SYSTEM, 
					"parentalControlServiceApplicationRestricted"
				), 
				0,
				CHANNEL_0,
				SYSTEM
			);

			::parentalControlRestrictionDispatched = true;
		}
	}
}

void DispatchProControllerConnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(pro_controller_connected->size() <= channel) {
		pro_controller_connected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(input->HasProController(channel)) {
		if(!(*pro_controller_connected)[channel]) {
			(*pro_controller_connected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 1.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerConnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					PRO, 
					(WiiUSystemChannel)channel,
					DEFAULT,
					"controllerConnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				PRO,
				DEFAULT
			);
		} 
	} else {
		(*pro_controller_connected)[channel] = false;
	}
}

void DispatchProControllerDisconnect(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Expand the array to however many remotes we have
	while(pro_controller_disconnected->size() <= channel) {
		pro_controller_disconnected->push_back(true);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	// If the state of the remote has changed send the appropriate event
	if(!input->HasProController(channel)) {
		if(!(*pro_controller_disconnected)[channel]) {
			(*pro_controller_disconnected)[channel] = true;
			WiiUScalarData scalar_data;
			scalar_data.channel = channel;
			scalar_data.s = 0.0f;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"controllerDisconnected",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					PRO, 
					(WiiUSystemChannel)channel,
					DEFAULT, 
					"controllerDisconnected"
				), 
				(void *)&scalar_data,
				(WiiUSystemChannel)channel,
				PRO,
				DEFAULT
			);
		}
	} else {
		(*pro_controller_disconnected)[channel] = false;
	}
}

void DispatchProControllerBatteryLevelChanged(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	
	if(!input->HasProController(channel)) return;

	while(pro_controller_battery_levels->size() <= channel) {
		pro_controller_battery_levels->push_back(input->GetRemoteBatteryLevel(channel));
	}

	if(!input->ProControllerEnabled(channel)) {
		return;
	}

	if(input->GetRemoteBatteryLevel(channel) != (*pro_controller_battery_levels)[channel]) {
		(*pro_controller_battery_levels)[channel] = input->GetRemoteBatteryLevel(channel);
		WiiUScalarData scalar_data;
		scalar_data.channel = channel;
		scalar_data.s = 0.0f;

		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
			"batteryLevelChange",
			JSSystemEventDispatcher::GetInstance()->GetEventString(
				PRO, 
				(WiiUSystemChannel)channel,
				DEFAULT,
				"batteryLevelChange"
			), 
			(void *)&scalar_data,
			(WiiUSystemChannel)channel,
			PRO,
			DEFAULT
		);
	}
}

void DispatchProControllerTrigger(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasProController(channel)) {
		if(!input->ProControllerEnabled(channel)) return;
		
		unsigned trigger_data = input->ProControllerWasTriggered(channel, ::buffer);

		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = trigger_data;

		unsigned buttonBitFlag = 1;
		unsigned buttonBitMax = 1 << 31;

		//This loop will fire events for each possible button
		while(buttonBitFlag <= trigger_data)
		{
			if(buttonBitFlag & trigger_data)
			{
				key_data.key_state = buttonBitFlag;
				JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
					"press",
					JSSystemEventDispatcher::GetInstance()->GetEventString(
						PRO, 
						(WiiUSystemChannel)channel,
						BUTTON,
						"press"
					), 
					(void *)&key_data,
					(WiiUSystemChannel)channel,
					PRO,
					BUTTON
				);
			}

			if(buttonBitFlag == buttonBitMax)
				break;

			buttonBitFlag = buttonBitFlag << 1;
		}
	}
}

void DispatchProControllerRelease(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();
	if(::buffer < remote_reads_behind && input->HasProController(channel)) {
		if(!input->ProControllerEnabled(channel)) {
			return;
		}

		unsigned release_data = input->ProControllerWasReleased(channel, ::buffer);
		
		WiiUKeyData key_data;
		key_data.channel = channel;
		key_data.key_state = release_data;
			
		unsigned buttonBitFlag = 1;
		unsigned buttonBitMax = 1 << 31;

		//This loop will fire events for each possible button
		while(buttonBitFlag <= release_data)
		{
			if(buttonBitFlag & release_data)
			{
				key_data.key_state = buttonBitFlag;
				JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
					"release",
					JSSystemEventDispatcher::GetInstance()->GetEventString(
						PRO, 
						(WiiUSystemChannel)channel,
						BUTTON,
						"release"
					), 
					(void *)&key_data,
					(WiiUSystemChannel)channel,
					PRO,
					BUTTON
				);
			}
			if(buttonBitFlag == buttonBitMax)
				break;
			buttonBitFlag = buttonBitFlag << 1;
		}
		
	}
}

void DispatchProControllerLStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;

	while(pro_controller_lstick_was_active->size() <= channel) {
		pro_controller_lstick_was_active->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasProController(channel)) {
		if(!input->ProControllerEnabled(channel)){
			return;
		}

		float2 lstick_data = input->ProControllerLStick(channel);

		if(lstick_data.x || lstick_data.y || (*pro_controller_lstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*pro_controller_lstick_was_active)[channel] = lstick_data.x || lstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = lstick_data.x;
			vector2_data.y = lstick_data.y;
			
			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					PRO, 
					(WiiUSystemChannel)channel,
					LSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				PRO,
				LSTICK
			);
		}
	}
}

void DispatchProControllerRStick(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;

	while(pro_controller_rstick_was_active->size() <= channel) {
		pro_controller_rstick_was_active->push_back(false);
	}

	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < remote_reads_behind && input->HasProController(channel)) {
		if(!input->ProControllerEnabled(channel)) {
			return;
		}

		float2 rstick_data = input->ProControllerRStick(channel);

		if(rstick_data.x || rstick_data.y || (*pro_controller_rstick_was_active)[channel]) {
			//this allows the system to fire a final move event when the stick is released
			(*pro_controller_rstick_was_active)[channel] = rstick_data.x || rstick_data.y;
			WiiUVector2Data vector2_data;
			vector2_data.channel = channel;
			vector2_data.x = rstick_data.x;
			vector2_data.y = rstick_data.y;

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"move",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					PRO, 
					(WiiUSystemChannel)channel,
					RSTICK,
					"move"
				), 
				(void *)&vector2_data,
				(WiiUSystemChannel)channel,
				PRO,
				RSTICK
			);
		}
	}
}

void DispatchProControllerTrigOrRel(signed channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	LibPlatform::Input *input = LibPlatform::Input::GetInstance();

	if(::buffer < gamepad_reads_behind && input->HasProController(channel)) {
		if(!input->ProControllerEnabled(channel)) {
			return;
		}

		unsigned trigger_data = input->ProControllerTriggered(channel);
		unsigned release_data = input->ProControllerReleased(channel);

		if((trigger_data | release_data) & (REMOTE_UC_BUTTON_LEFT | REMOTE_UC_BUTTON_UP | REMOTE_UC_BUTTON_RIGHT | REMOTE_UC_BUTTON_DOWN)) {
			WiiUVector2Data key_data;
			key_data.channel = channel;
			
			unsigned press_data = input->ProControllerPressed(channel) & (REMOTE_UC_BUTTON_LEFT | REMOTE_UC_BUTTON_UP | REMOTE_UC_BUTTON_RIGHT | REMOTE_UC_BUTTON_DOWN);

			if(press_data == REMOTE_UC_BUTTON_LEFT) 								{key_data.x = -1; key_data.y =  0;}
			else if(press_data == REMOTE_UC_BUTTON_UP) 								{key_data.x =  0; key_data.y =  1;}
			else if(press_data == REMOTE_UC_BUTTON_RIGHT) 							{key_data.x =  1; key_data.y =  0;}
			else if(press_data == REMOTE_UC_BUTTON_DOWN) 							{key_data.x =  0; key_data.y = -1;}
			else if(press_data == (REMOTE_UC_BUTTON_LEFT | REMOTE_UC_BUTTON_UP)) 	{key_data.x = -1; key_data.y =  1;}
			else if(press_data == (REMOTE_UC_BUTTON_UP | REMOTE_UC_BUTTON_RIGHT)) 	{key_data.x =  1; key_data.y =  1;}
			else if(press_data == (REMOTE_UC_BUTTON_RIGHT | REMOTE_UC_BUTTON_DOWN)) {key_data.x =  1; key_data.y = -1;}
			else if(press_data == (REMOTE_UC_BUTTON_DOWN | REMOTE_UC_BUTTON_LEFT)) 	{key_data.x = -1; key_data.y = -1;}
			else 																	{key_data.x =  0; key_data.y =  0;}

			JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
				"directionChange",
				JSSystemEventDispatcher::GetInstance()->GetEventString(
					PRO, 
					(WiiUSystemChannel)channel,
					DIRECTION,
					"directionChange"
				), 
				(void *)&key_data,
				(WiiUSystemChannel)channel,
				PRO,
				DIRECTION
			);
		}
	}
}

// TODO: This shouldn't be necessary. Investigate!
void DispatchNoOp(signed) {}

namespace {
	JSSystemEventDispatcher *instance = 0;
};

//////////////////////////////////////////////////////////////////////////
///	Name	: HashString_DJB2
///
///	Purpose	: Uses the DJB2 algorithm for string hashing to return a
///				hash key for the given string. The size of the returned
///				hash key can be specified using the TYPE template
///				parameter.
///
///	NOTE	: It has been verified that none of the strings this is
///				hashing hash to the same value. This allows us to use
///				simple integers as the event string's map key, rather
///				than doing string compares.
/////////////////////////////////////////////////////////////////////////
unsigned int HashString_DJB2(const char * vString)
{
	unsigned int Hash = 5381;
	const unsigned char * CurrentValue = (const unsigned char*)vString;

	//If no string was given, we can't hash it
	if (!vString)
		return 0;

	//Hash the string
	while (*CurrentValue)
	{
		Hash = ((Hash << 5) + Hash) + *CurrentValue;
		CurrentValue++;
	}

	//Return the hash
	//LibPlatform::OS::Report("Hashem: %s = 0x%X\n", vString, Hash);
	return Hash;
}

JSSystemEventDispatcher::JSSystemEventDispatcher(void) : JSEventDispatcher("SystemDispatcher"), crashing(false), crash_callback(0) {

	pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&eventListenerMapMutex, &attr);
	pthread_mutex_init(&addListMutex, &attr);

	// REMOTE unique event strings
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("REMOTE_CONNECT0", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("REMOTE_DISCONNECT0", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("extensionControllerAdded")] =			KeyDispatcherPair("REMOTE_EXTADDED0", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("extensionControllerRemoved")] =		KeyDispatcherPair("REMOTE_EXTREMOVED0", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("REMOTE_BATTERYCHANGED0", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("mplsAdded")] =				KeyDispatcherPair("REMOTE_MPLSADDED0", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_0][DEFAULT][HashString_DJB2("mplsRemoved")] =				KeyDispatcherPair("REMOTE_MPLSREMOVED0", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_0][BUTTON][HashString_DJB2("press")] =						KeyDispatcherPair("REMOTE_TRIGGERED0", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_0][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("REMOTE_RELEASED0", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_0][DPD][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_MOVED0", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_0][ACCELEROMETER][HashString_DJB2("update")] =				KeyDispatcherPair("REMOTE_ACCEL0", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_0][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL0", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_0][GYRO][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_GYRO0", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_0][GYRO][HashString_DJB2("calibrationSuccess")] =			KeyDispatcherPair("REMOTE_GYROCALIBRATION0", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_0][GYRO][HashString_DJB2("calibrationFail")] =				KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL0", DispatchRemoteGyroCalibrationFailed);
	
																					 
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("REMOTE_CONNECT1", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("REMOTE_DISCONNECT1", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("extensionControllerAdded")] =			KeyDispatcherPair("REMOTE_EXTADDED1", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("extensionControllerRemoved")] =		KeyDispatcherPair("REMOTE_EXTREMOVED1", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("REMOTE_BATTERYCHANGED1", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("mplsAdded")] =				KeyDispatcherPair("REMOTE_MPLSADDED1", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_1][DEFAULT][HashString_DJB2("mplsRemoved")] =				KeyDispatcherPair("REMOTE_MPLSREMOVED1", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_1][BUTTON][HashString_DJB2("press")] =						KeyDispatcherPair("REMOTE_TRIGGERED1", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_1][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("REMOTE_RELEASED1", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_1][DPD][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_MOVED1", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_1][ACCELEROMETER][HashString_DJB2("update")] =				KeyDispatcherPair("REMOTE_ACCEL1", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_1][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL1", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_1][GYRO][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_GYRO1", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_1][GYRO][HashString_DJB2("calibrationSuccess")] =			KeyDispatcherPair("REMOTE_GYROCALIBRATION1", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_1][GYRO][HashString_DJB2("calibrationFail")] =				KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL1", DispatchRemoteGyroCalibrationFailed);
																					 
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("REMOTE_CONNECT2", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("REMOTE_DISCONNECT2", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("extensionControllerAdded")] =			KeyDispatcherPair("REMOTE_EXTADDED2", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("extensionControllerRemoved")] =		KeyDispatcherPair("REMOTE_EXTREMOVED2", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("REMOTE_BATTERYCHANGED2", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("mplsAdded")] =				KeyDispatcherPair("REMOTE_MPLSADDED2", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_2][DEFAULT][HashString_DJB2("mplsRemoved")] =				KeyDispatcherPair("REMOTE_MPLSREMOVED2", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_2][BUTTON][HashString_DJB2("press")] =						KeyDispatcherPair("REMOTE_TRIGGERED2", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_2][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("REMOTE_RELEASED2", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_2][DPD][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_MOVED2", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_2][ACCELEROMETER][HashString_DJB2("update")] =				KeyDispatcherPair("REMOTE_ACCEL2", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_2][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL2", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_2][GYRO][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_GYRO2", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_2][GYRO][HashString_DJB2("calibrationSuccess")] =			KeyDispatcherPair("REMOTE_GYROCALIBRATION2", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_2][GYRO][HashString_DJB2("calibrationFail")] =				KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL2", DispatchRemoteGyroCalibrationFailed);
																					 
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("REMOTE_CONNECT3", DispatchRemoteConnect);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("REMOTE_DISCONNECT3", DispatchRemoteDisconnect);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("extensionControllerAdded")] =			KeyDispatcherPair("REMOTE_EXTADDED3", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("extensionControllerRemoved")] =		KeyDispatcherPair("REMOTE_EXTREMOVED3", DispatchRemoteExtensionAddedRemoved);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("REMOTE_BATTERYCHANGED3", DispatchRemoteBatteryLevelChanged);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("mplsAdded")] =				KeyDispatcherPair("REMOTE_MPLSADDED3", DispatchRemoteMplsAttached);
	event_strings[REMOTE][CHANNEL_3][DEFAULT][HashString_DJB2("mplsRemoved")] =				KeyDispatcherPair("REMOTE_MPLSREMOVED3", DispatchRemoteMplsRemoved);
	event_strings[REMOTE][CHANNEL_3][BUTTON][HashString_DJB2("press")] =						KeyDispatcherPair("REMOTE_TRIGGERED3", DispatchRemoteTrigger);
	event_strings[REMOTE][CHANNEL_3][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("REMOTE_RELEASED3", DispatchRemoteRelease);
	event_strings[REMOTE][CHANNEL_3][DPD][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_MOVED3", DispatchRemotePointer);
	event_strings[REMOTE][CHANNEL_3][ACCELEROMETER][HashString_DJB2("update")] =				KeyDispatcherPair("REMOTE_ACCEL3", DispatchRemoteAcceleration);
	event_strings[REMOTE][CHANNEL_3][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_TRIG_OR_REL3", DispatchRemoteTrigOrRel);
	event_strings[REMOTE][CHANNEL_3][GYRO][HashString_DJB2("update")] =						KeyDispatcherPair("REMOTE_GYRO3", DispatchRemoteGyro);
	event_strings[REMOTE][CHANNEL_3][GYRO][HashString_DJB2("calibrationSuccess")] =			KeyDispatcherPair("REMOTE_GYROCALIBRATION3", DispatchRemoteGyroCalibration);
	event_strings[REMOTE][CHANNEL_3][GYRO][HashString_DJB2("calibrationFail")] =				KeyDispatcherPair("REMOTE_GYROCALIBRATIONFAIL3", DispatchRemoteGyroCalibrationFailed);

	// Classic controller strings
	event_strings[REMOTE_CL][CHANNEL_0][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_CL_TRIGGERED0", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_0][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_CL_RELEASED0", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_0][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_LSTICK0", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_0][RSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_RSTICK0", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_0][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL0", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_1][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_CL_TRIGGERED1", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_1][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_CL_RELEASED1", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_1][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_LSTICK1", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_1][RSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_RSTICK1", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_1][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL1", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_2][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_CL_TRIGGERED2", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_2][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_CL_RELEASED2", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_2][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_LSTICK2", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_2][RSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_RSTICK2", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_2][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL2", DispatchClassicTrigOrRel);
																		
	event_strings[REMOTE_CL][CHANNEL_3][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_CL_TRIGGERED3", DispatchClassicTrigger);
	event_strings[REMOTE_CL][CHANNEL_3][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_CL_RELEASED3", DispatchClassicRelease);
	event_strings[REMOTE_CL][CHANNEL_3][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_LSTICK3", DispatchClassicLStick);
	event_strings[REMOTE_CL][CHANNEL_3][RSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_CL_RSTICK3", DispatchClassicRStick);
	event_strings[REMOTE_CL][CHANNEL_3][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("REMOTE_CL_TRIG_OR_REL3", DispatchClassicTrigOrRel);
																		
	// Nun-chuck controller event strings					
	event_strings[REMOTE_FS][CHANNEL_0][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_FS_TRIGGERED0", DispatchNunChuckPress);
	event_strings[REMOTE_FS][CHANNEL_0][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_FS_RELEASED0", DispatchNunChuckRelease);
	event_strings[REMOTE_FS][CHANNEL_0][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_FS_STICK0", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_0][ACCELEROMETER][HashString_DJB2("update")] =			KeyDispatcherPair("REMOTE_FS_ACCEL0", DispatchNunChuckAccel);
		
	event_strings[REMOTE_FS][CHANNEL_1][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_FS_TRIGGERED1", DispatchNunChuckPress);
	event_strings[REMOTE_FS][CHANNEL_1][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_FS_RELEASED1", DispatchNunChuckRelease);
	event_strings[REMOTE_FS][CHANNEL_1][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_FS_STICK1", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_1][ACCELEROMETER][HashString_DJB2("update")] =			KeyDispatcherPair("REMOTE_FS_ACCEL1", DispatchNunChuckAccel);
	
	event_strings[REMOTE_FS][CHANNEL_2][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_FS_TRIGGERED2", DispatchNunChuckPress);
	event_strings[REMOTE_FS][CHANNEL_2][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_FS_RELEASED2", DispatchNunChuckRelease);
	event_strings[REMOTE_FS][CHANNEL_2][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_FS_STICK2", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_2][ACCELEROMETER][HashString_DJB2("update")] =			KeyDispatcherPair("REMOTE_FS_ACCEL2", DispatchNunChuckAccel);
		
	event_strings[REMOTE_FS][CHANNEL_3][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("REMOTE_FS_TRIGGERED3", DispatchNunChuckPress);
	event_strings[REMOTE_FS][CHANNEL_3][BUTTON][HashString_DJB2("release")] =				KeyDispatcherPair("REMOTE_FS_RELEASED3", DispatchNunChuckRelease);
	event_strings[REMOTE_FS][CHANNEL_3][LSTICK][HashString_DJB2("move")] =					KeyDispatcherPair("REMOTE_FS_STICK3", DispatchNunChuckStick);
	event_strings[REMOTE_FS][CHANNEL_3][ACCELEROMETER][HashString_DJB2("update")] =			KeyDispatcherPair("REMOTE_FS_ACCEL3", DispatchNunChuckAccel);

	// GAMEPAD unique event strings
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("GAMEPAD_CONNECT0", DispatchGamePadConnect);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("GAMEPAD_DISCONNECT0", DispatchGamePadDisconnect);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("GAMEPAD_BATTERYCHANGED0", DispatchGamePadBatteryLevelChanged);
	event_strings[GAMEPAD][CHANNEL_0][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("GAMEPAD_TRIGGERED0", DispatchGamePadTrigger);
	event_strings[GAMEPAD][CHANNEL_0][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("GAMEPAD_RELEASED0", DispatchGamePadRelease);
	event_strings[GAMEPAD][CHANNEL_0][TOUCH][HashString_DJB2("update")] =					KeyDispatcherPair("GAMEPAD_MOVED0", DispatchGamePadPointer);
	event_strings[GAMEPAD][CHANNEL_0][TOUCH][HashString_DJB2("touchStart")] =				KeyDispatcherPair("GAMEPAD_TOUCHSTART0", DispatchGamePadTouchStart);
	event_strings[GAMEPAD][CHANNEL_0][TOUCH][HashString_DJB2("touchEnd")] =					KeyDispatcherPair("GAMEPAD_TOUCHEND0", DispatchGamePadTouchEnd);
	event_strings[GAMEPAD][CHANNEL_0][ACCELEROMETER][HashString_DJB2("update")] =			KeyDispatcherPair("GAMEPAD_ACCEL0", DispatchGamePadAcceleration);
	event_strings[GAMEPAD][CHANNEL_0][LSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("GAMEPAD_LSTICK0", DispatchGamePadLStick);
	event_strings[GAMEPAD][CHANNEL_0][RSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("GAMEPAD_RSTICK0", DispatchGamePadRStick);
	event_strings[GAMEPAD][CHANNEL_0][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("GAMEPAD_TRIG_OR_REL0", DispatchGamePadTrigOrRel);
	event_strings[GAMEPAD][CHANNEL_0][GYRO][HashString_DJB2("update")] =						KeyDispatcherPair("GAMEPAD_GYRO0", DispatchGamePadGyro);
	event_strings[GAMEPAD][CHANNEL_0][GYRO][HashString_DJB2("calibrationSuccess")] =			KeyDispatcherPair("GAMEPAD_GYROCALIBRATION0", DispatchGamePadGyroCalibration);
	event_strings[GAMEPAD][CHANNEL_0][GYRO][HashString_DJB2("calibrationFail")] =			KeyDispatcherPair("GAMEPAD_GYROCALIBRATIONFAIL0", DispatchGamePadGyroCalibrationFailed);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("ready")] =					KeyDispatcherPair("CAM_READY0", DispatchCAMReady);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("open")] =					KeyDispatcherPair("CAM_STREAM_OPEN0", DispatchCAMStreamOpen);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("close")] =					KeyDispatcherPair("CAM_STREAM_CLOSE0", DispatchCAMStreamClose);
	event_strings[GAMEPAD][CHANNEL_0][DEFAULT][HashString_DJB2("exit")] =					KeyDispatcherPair("CAM_EXIT0", DispatchCAMExit);
    event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("ready")] =					    KeyDispatcherPair("MIC_READY0", DispatchMICReady);
	event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("open")] =					    KeyDispatcherPair("MIC_OPEN0", DispatchMICOpen);
    event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("update")] =					    KeyDispatcherPair("MIC_STREAM_UPDATE0", DispatchMICStreamUpdate);
	event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("exit")] =					    KeyDispatcherPair("MIC_EXIT0", DispatchMICExit);
	event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("close")] =					    KeyDispatcherPair("MIC_CLOSE0", DispatchMICClose);
	event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("connected")] =				    KeyDispatcherPair("MIC_CONNECTED0", DispatchMICConnectedUpdate);
	event_strings[GAMEPAD][CHANNEL_0][MIC][HashString_DJB2("disconnected")] =			    KeyDispatcherPair("MIC_DISCONNECTED0", DispatchMICSDisconnectedClose);

	// File system strings
	event_strings[IO][CHANNEL_0][DEFAULT][HashString_DJB2("readComplete")] =					KeyDispatcherPair("IO_READ_COMPLETE", DispatchNoOp);
	event_strings[IO][CHANNEL_0][DEFAULT][HashString_DJB2("saveComplete")] =					KeyDispatcherPair("IO_SAVE_COMPLETE", DispatchNoOp);
	event_strings[IO][CHANNEL_0][DEFAULT][HashString_DJB2("error")] =						KeyDispatcherPair("IO_ERROR", DispatchNoOp);

	// Account strings
	event_strings[ACT][CHANNEL_0][DEFAULT][HashString_DJB2("accountServiceToken")] =			KeyDispatcherPair("ACCOUNT_SERVICE_TOKEN_SUCCESS", DispatchNoOp);
	event_strings[ACT][CHANNEL_0][DEFAULT][HashString_DJB2("accountServiceTokenError")] =	KeyDispatcherPair("ACCOUNT_SERVICE_TOKEN_FAIL", DispatchNoOp);
	event_strings[ACT][CHANNEL_0][DEFAULT][HashString_DJB2("accountEcServiceToken")] =		KeyDispatcherPair("ACCOUNT_EC_SERVICE_TOKEN_SUCCESS", DispatchNoOp);
	event_strings[ACT][CHANNEL_0][DEFAULT][HashString_DJB2("accountEcServiceTokenError")] =	KeyDispatcherPair("ACCOUNT_EC_SERVICE_TOKEN_FAIL", DispatchNoOp);

	// Network strings
	event_strings[NETWORK][CHANNEL_0][DEFAULT][HashString_DJB2("networkConnected")] =		KeyDispatcherPair("NETWORK_CONNECTED", DispatchNoOp);
	event_strings[NETWORK][CHANNEL_0][DEFAULT][HashString_DJB2("networkDisconnected")] =		KeyDispatcherPair("NETWORK_DISCONNECTED", DispatchNoOp);
	event_strings[NETWORK][CHANNEL_0][DEFAULT][HashString_DJB2("networkFailedToConnect")] =	KeyDispatcherPair("NETWORK_FAILED_TO_CONNECT", DispatchNoOp);

	// Software Keyboard strings
	event_strings[SOFTWAREKEYBOARD][CHANNEL_0][DEFAULT][HashString_DJB2("animationStart")] =		KeyDispatcherPair("ANIMATION_START", DispatchNoOp);
	event_strings[SOFTWAREKEYBOARD][CHANNEL_0][DEFAULT][HashString_DJB2("animationEnd")] =		KeyDispatcherPair("ANIMATION_END", DispatchNoOp);

    // Error strings
	event_strings[ERROR_WINDOW][CHANNEL_0][DEFAULT][HashString_DJB2("errorWindowClose")] =	KeyDispatcherPair("ERROR_WINDOW_CLOSE", DispatchNoOp);

	// System event strings
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("error")] =					KeyDispatcherPair("SYSTEM_ERROR", DispatchErrors);
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("usbKeyboardConnected")] =		KeyDispatcherPair("USBKEYBOARD_CONNECTED", DispatchNoOp);
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("usbKeyboardDisconnected")] =	KeyDispatcherPair("USBKEYBOARD_DISCONNECTED", DispatchNoOp);
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("crash")] =					KeyDispatcherPair("SYSTEM_CRASH", DispatchCrash);
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("parentalControlServiceApplicationRestricted")] =	KeyDispatcherPair("PARENTAL_CONTROL_SERVICE_APP_RESTRICTED", DispatchParentalControlServiceApplicationRestricted);
	event_strings[SYSTEM][CHANNEL_0][DEFAULT][HashString_DJB2("flushStorageComplete")] =	KeyDispatcherPair("FLUSH_STORAGE_COMPLETE", DispatchNoOp);

	// Display event strings
	//CHANNEL_0 == tv
	event_strings[DISPLAY][CHANNEL_0][DEFAULT][HashString_DJB2("initialized")] =				KeyDispatcherPair("DISPLAY_INIT0", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT][HashString_DJB2("load")] =					KeyDispatcherPair("DISPLAY_LOAD0", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT][HashString_DJB2("uninitialized")] =			KeyDispatcherPair("DISPLAY_UNINIT0", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_0][DEFAULT][HashString_DJB2("ready")] =					KeyDispatcherPair("DISPLAY_READY0", DispatchNoOp);

	//CHANNEL_1 == GAMEPAD_0
	event_strings[DISPLAY][CHANNEL_1][DEFAULT][HashString_DJB2("initialized")] =				KeyDispatcherPair("DISPLAY_INIT1", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT][HashString_DJB2("load")] =					KeyDispatcherPair("DISPLAY_LOAD1", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT][HashString_DJB2("uninitialized")] =			KeyDispatcherPair("DISPLAY_UNINIT1", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_1][DEFAULT][HashString_DJB2("ready")] =					KeyDispatcherPair("DISPLAY_READY1", DispatchNoOp);

	//CHANNEL_2 == GAMEPAD_1
	event_strings[DISPLAY][CHANNEL_2][DEFAULT][HashString_DJB2("initialized")] =				KeyDispatcherPair("DISPLAY_INIT2", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT][HashString_DJB2("load")] =					KeyDispatcherPair("DISPLAY_LOAD2", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT][HashString_DJB2("uninitialized")] =			KeyDispatcherPair("DISPLAY_UNINIT2", DispatchNoOp);
	event_strings[DISPLAY][CHANNEL_2][DEFAULT][HashString_DJB2("ready")] =					KeyDispatcherPair("DISPLAY_READY2", DispatchNoOp);

	// Balance board event strings
	event_strings[BALANCEBOARD][CHANNEL_0][WEIGHT][HashString_DJB2("weightChange")] =				KeyDispatcherPair("BALANCEBOARD_READ0", DispatchBalanceBoardInfo);
	event_strings[BALANCEBOARD][CHANNEL_1][WEIGHT][HashString_DJB2("weightChange")] =				KeyDispatcherPair("BALANCEBOARD_READ1", DispatchBalanceBoardInfo);
	event_strings[BALANCEBOARD][CHANNEL_2][WEIGHT][HashString_DJB2("weightChange")] =				KeyDispatcherPair("BALANCEBOARD_READ2", DispatchBalanceBoardInfo);
	event_strings[BALANCEBOARD][CHANNEL_3][WEIGHT][HashString_DJB2("weightChange")] =				KeyDispatcherPair("BALANCEBOARD_READ3", DispatchBalanceBoardInfo);
	event_strings[BALANCEBOARD][CHANNEL_3][DEFAULT][HashString_DJB2("weightChange")] =				KeyDispatcherPair("BALANCEBOARD_READ", DispatchCorrectedBalanceBoardInfo);
	event_strings[BALANCEBOARD][CHANNEL_3][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("BALANCE_BOARD_CONNECT", DispatchBBoardConnect);
	event_strings[BALANCEBOARD][CHANNEL_3][DEFAULT][HashString_DJB2("controllerDisconnected")] =		KeyDispatcherPair("BALANCE_BOARD_DISCONNECT", DispatchBBoardDisconnect);
	event_strings[BALANCEBOARD][CHANNEL_3][DEFAULT][HashString_DJB2("batteryLevelChange")] =			KeyDispatcherPair("BALANCE_BOARD_BATTERYCHANGED", DispatchBalanceBoardBatteryLevelChanged);

	// Pro Controller event strings
	event_strings[PRO][CHANNEL_0][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("PRO_CONNECT0", DispatchProControllerConnect);
	event_strings[PRO][CHANNEL_0][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("PRO_DISCONNECT0", DispatchProControllerDisconnect);
	event_strings[PRO][CHANNEL_0][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("PRO_BATTERYCHANGED0", DispatchProControllerBatteryLevelChanged);
	event_strings[PRO][CHANNEL_0][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("PRO_TRIGGERED0", DispatchProControllerTrigger);
	event_strings[PRO][CHANNEL_0][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("PRO_RELEASED0", DispatchProControllerRelease);
	event_strings[PRO][CHANNEL_0][LSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_LSTICK0", DispatchProControllerLStick);
	event_strings[PRO][CHANNEL_0][RSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_RSTICK0", DispatchProControllerRStick);
	event_strings[PRO][CHANNEL_0][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("PRO_TRIG_OR_REL0", DispatchProControllerTrigOrRel);

	event_strings[PRO][CHANNEL_1][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("PRO_CONNECT1", DispatchProControllerConnect);
	event_strings[PRO][CHANNEL_1][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("PRO_DISCONNECT1", DispatchProControllerDisconnect);
	event_strings[PRO][CHANNEL_1][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("PRO_BATTERYCHANGED1", DispatchProControllerBatteryLevelChanged);
	event_strings[PRO][CHANNEL_1][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("PRO_TRIGGERED1", DispatchProControllerTrigger);
	event_strings[PRO][CHANNEL_1][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("PRO_RELEASED1", DispatchProControllerRelease);
	event_strings[PRO][CHANNEL_1][LSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_LSTICK1", DispatchProControllerLStick);
	event_strings[PRO][CHANNEL_1][RSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_RSTICK1", DispatchProControllerRStick);
	event_strings[PRO][CHANNEL_1][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("PRO_TRIG_OR_REL1", DispatchProControllerTrigOrRel);

	event_strings[PRO][CHANNEL_2][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("PRO_CONNECT2", DispatchProControllerConnect);
	event_strings[PRO][CHANNEL_2][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("PRO_DISCONNECT2", DispatchProControllerDisconnect);
	event_strings[PRO][CHANNEL_2][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("PRO_BATTERYCHANGED2", DispatchProControllerBatteryLevelChanged);
	event_strings[PRO][CHANNEL_2][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("PRO_TRIGGERED2", DispatchProControllerTrigger);
	event_strings[PRO][CHANNEL_2][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("PRO_RELEASED2", DispatchProControllerRelease);
	event_strings[PRO][CHANNEL_2][LSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_LSTICK2", DispatchProControllerLStick);
	event_strings[PRO][CHANNEL_2][RSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_RSTICK2", DispatchProControllerRStick);
	event_strings[PRO][CHANNEL_2][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("PRO_TRIG_OR_REL2", DispatchProControllerTrigOrRel);

	event_strings[PRO][CHANNEL_3][DEFAULT][HashString_DJB2("controllerConnected")] =		KeyDispatcherPair("PRO_CONNECT3", DispatchProControllerConnect);
	event_strings[PRO][CHANNEL_3][DEFAULT][HashString_DJB2("controllerDisconnected")] =	KeyDispatcherPair("PRO_DISCONNECT3", DispatchProControllerDisconnect);
	event_strings[PRO][CHANNEL_3][DEFAULT][HashString_DJB2("batteryLevelChange")] =		KeyDispatcherPair("PRO_BATTERYCHANGED3", DispatchProControllerBatteryLevelChanged);
	event_strings[PRO][CHANNEL_3][BUTTON][HashString_DJB2("press")] =					KeyDispatcherPair("PRO_TRIGGERED3", DispatchProControllerTrigger);
	event_strings[PRO][CHANNEL_3][BUTTON][HashString_DJB2("release")] =					KeyDispatcherPair("PRO_RELEASED3", DispatchProControllerRelease);
	event_strings[PRO][CHANNEL_3][LSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_LSTICK3", DispatchProControllerLStick);
	event_strings[PRO][CHANNEL_3][RSTICK][HashString_DJB2("move")] =						KeyDispatcherPair("PRO_RSTICK3", DispatchProControllerRStick);
	event_strings[PRO][CHANNEL_3][DIRECTION][HashString_DJB2("directionChange")] =		KeyDispatcherPair("PRO_TRIG_OR_REL3", DispatchProControllerTrigOrRel);

	g_DestructingJSSystemEventDispatcher = false;
}

JSSystemEventDispatcher::KeyDispatcherPair::~KeyDispatcherPair()
{
	dispatcher = 0;
}

JSSystemEventDispatcher::~JSSystemEventDispatcher(void) {
	LibPlatform::OS::Report("Deleting the event strings (this=0x%X)\n", this);

	pthread_mutex_destroy(&addListMutex);

	std::map<std::string, KeyDispatcherPair>::iterator MapIterator;

	//Sadly, this seems to be the only way to stop the map from crashing when it clears.
	//Remove all the event strings one by one!
	for (int i = 0; i < DEVICE_COUNT; i++)
	{
		for (int j = 0; j < CHANNEL_COUNT; j++)
		{
			for (int k = 0; k < CONTROL_COUNT; k++)
			{
				while (event_strings[i][j][k].size())
					event_strings[i][j][k].erase(event_strings[i][j][k].begin());
			}
		}
		
	}
	

	LibPlatform::OS::Report("Event strings deleted\n");
}

JSSystemEventDispatcher *JSSystemEventDispatcher::GetInstance(void) {
	if(!::instance) {
		::instance = new JSSystemEventDispatcher();
	}

	return ::instance;
}

void JSSystemEventDispatcher::DestroyInstance(void) {
	if(::instance) {
		g_DestructingJSSystemEventDispatcher = true;
		delete ::instance;
		::instance = 0;
		g_DestructingJSSystemEventDispatcher = false;
	}
}

bool JSSystemEventDispatcher::AddDefaultEventListener(JSEventListener *js_listener, std::string event_type) {
	pthread_mutex_lock(&addListMutex);
	listener_add_list.push_back(ListenerEventPair(js_listener, event_type));
	pthread_mutex_unlock(&addListMutex);
	return true;
}

std::string JSSystemEventDispatcher::GetEventString(WiiUSystemDevice device, WiiUSystemChannel channel, WiiUSystemControl control, std::string event_type) {
	return event_strings[device][channel][control][HashString_DJB2(event_type.c_str())].key;
}

std::string JSSystemEventDispatcher::GetEventString(WiiUSystemDevice device, std::string event_type) {
	return this->GetEventString(device, CHANNEL_0, DEFAULT, event_type);
}

bool JSSystemEventDispatcher::AddEventListener(
	JSEventListener *js_listener, 
	WiiUSystemDevice device, 
	WiiUSystemChannel channel, 
	WiiUSystemControl control,
	std::string event_type) 
{
	KeyDispatcherPair pair = event_strings[device][channel][control][HashString_DJB2(event_type.c_str())];
	if(!pair.dispatcher) return false;
	pthread_mutex_lock(&addListMutex);
	this->add_list.push_back(ListenerEventFunctorTuple(js_listener, pair.key, DispatchFunctor(pair.dispatcher, channel, pair.key)));
	pthread_mutex_unlock(&addListMutex);
	return true;
}

bool JSSystemEventDispatcher::AddEventListener(
	JSEventListener *js_listener,
	WiiUSystemDevice device,
	std::string event_type) 
{
	return AddEventListener(js_listener, device, CHANNEL_0, DEFAULT, event_type);
}

unsigned int JSSystemEventDispatcher::HasEventListener(std::string event_type)
{
	typedef std::vector<ListenerEventFunctorTuple>::iterator M;
	pthread_mutex_lock(&addListMutex);
	for(M i = this->add_list.begin(); i != this->add_list.end(); ++i) {
		if(i->event_type == event_type)
		{
			pthread_mutex_unlock(&addListMutex);
			return 1 + JSEventDispatcher::HasEventListener(event_type);
		}
	}
	pthread_mutex_unlock(&addListMutex);

	return JSEventDispatcher::HasEventListener(event_type);
}

bool JSSystemEventDispatcher::HasEventListener(JSEventListener *js_listener, std::string event_type)
{
	typedef std::vector<ListenerEventFunctorTuple>::iterator M;
	pthread_mutex_lock(&addListMutex);
	for(M i = this->add_list.begin(); i != this->add_list.end(); ++i) {
		if(i->event_type == event_type && i->js_listener == js_listener)
		{
			pthread_mutex_unlock(&addListMutex);
			return true;
		}
	}
	pthread_mutex_unlock(&addListMutex);

	return JSEventDispatcher::HasEventListener(js_listener, event_type);
}

void JSSystemEventDispatcher::RemoveEventListener(JSEventListener *js_listener, std::string event_type) {
	
	pthread_mutex_lock(&eventListenerMapMutex);
	JSEventDispatcher::RemoveEventListener(js_listener, event_type);
	
	//if no listeners remain
	if(!JSEventDispatcher::HasEventListener(event_type))
	{
		//remove the matching functor
		typedef std::vector<DispatchFunctor>::iterator I;
		for(I i = this->dispatch_functors.begin(); i != this->dispatch_functors.end(); ++i)
		{
			if (i->event_type == event_type)
			{
				this->dispatch_functors.erase(i);
				break;
			}
		}
	}
	if(HasEventListener(js_listener, event_type))
	{
		pthread_mutex_lock(&addListMutex);
		
		std::vector<ListenerEventFunctorTuple>::iterator i = this->add_list.begin();
		while(i != this->add_list.end()) {
			if(i->event_type == event_type && i->js_listener == js_listener) {
				i = this->add_list.erase(i);
			} else i++;
		}

		pthread_mutex_unlock(&addListMutex);
	}
	pthread_mutex_unlock(&eventListenerMapMutex);
}

void JSSystemEventDispatcher::RemoveAllListeners(JSEventListener *js_listener)
{
	pthread_mutex_lock(&eventListenerMapMutex);
	JSEventDispatcher::RemoveAllListeners(js_listener);

	std::vector<ListenerEventFunctorTuple>::iterator i = this->add_list.begin();
	while(i != this->add_list.end()) {
		if(i->js_listener == js_listener) {
			i = this->add_list.erase(i);
		} else i++;
	}

	pthread_mutex_unlock(&addListMutex);
	pthread_mutex_unlock(&eventListenerMapMutex);
}

void JSSystemEventDispatcher::DispatchInputEvents(void) {
	if (g_DestructingJSSystemEventDispatcher) return;
	// Check to see if we're in a crash state.
	// If we are let the events dispatch then call the crash callback.
	// We check this here because we want to make sure it was set to true before we dispatch the events.
	bool call_crash_callback = this->crashing;

	// Figure out how far behind the main input thread we are
	Input *input = Input::GetInstance();
	// Add all listeners that don't use functors
	pthread_mutex_lock(&eventListenerMapMutex);
	pthread_mutex_lock(&addListMutex);
	typedef std::vector<ListenerEventPair>::iterator K;
	for(K k = this->listener_add_list.begin(); k != this->listener_add_list.end(); ++k) {
		JSEventDispatcher::AddEventListener(k->js_listener, k->event_type);
	}
	this->listener_add_list.clear();
	pthread_mutex_unlock(&addListMutex);

	// Its possible that a dispatch functor is added while dispatching an event.
	// So that we dont get a concurrent modification issue copy over pointers to the current list of functors
	// and iterate over that instead of the functor list.

	this->dispatch_functor_pointers.clear();
	typedef std::vector<ListenerEventFunctorTuple>::iterator M;
	typedef std::vector<DispatchFunctor>::iterator I;
	pthread_mutex_lock(&addListMutex);
	for(M i = this->add_list.begin(); i != this->add_list.end(); ++i) {
		
		if(!JSEventDispatcher::AddEventListener(i->js_listener, i->event_type)) continue;

		//We have to make sure that the Dispatch Functor isn't added to the list twice
		I dfp;
		for(dfp = this->dispatch_functors.begin(); dfp != this->dispatch_functors.end(); ++dfp) {
			//APW - 10/1/12: JSSystemEventDispatcher needs to check against event_name when adding listeners because some events use the same function 
			if(dfp->function == i->functor.function && dfp->event_type == i->functor.event_type) {
				break;
			}
		}
		//if the dispatch functor isn't already in the list, add it
		if(dfp == this->dispatch_functors.end())
		{
			this->dispatch_functors.push_back(i->functor);
		}
	}
	this->add_list.clear();
	pthread_mutex_unlock(&addListMutex);
	pthread_mutex_unlock(&eventListenerMapMutex);

	for(I i = this->dispatch_functors.begin(); i != this->dispatch_functors.end(); ++i) {
		this->dispatch_functor_pointers.push_back(&(*i));
	}

	::remote_reads_behind = 0;
	::gamepad_reads_behind = 0;

	unsigned cached_remote_count0 = input->HasCachedRemoteReads(0);
	unsigned cached_remote_count1 = input->HasCachedRemoteReads(1);
	unsigned cached_remote_count2 = input->HasCachedRemoteReads(2);
	unsigned cached_remote_count3 = input->HasCachedRemoteReads(3);
	unsigned cached_gamepad_count0 = input->HasCachedGamePadReads(0);

	if(input->HasRemoteController(0) || input->HasProController(0)) ::remote_reads_behind = std::max(cached_remote_count0, ::remote_reads_behind);
	if(input->HasRemoteController(1) || input->HasProController(1)) ::remote_reads_behind = std::max(cached_remote_count1, ::remote_reads_behind);
	if(input->HasRemoteController(2) || input->HasProController(2)) ::remote_reads_behind = std::max(cached_remote_count2, ::remote_reads_behind);
	if(input->HasRemoteController(3) || input->HasProController(3)) ::remote_reads_behind = std::max(cached_remote_count3, ::remote_reads_behind);
	if(input->HasGamePadController(0)) ::gamepad_reads_behind = std::max(cached_gamepad_count0, ::gamepad_reads_behind);

	// Dispatch events for every input update that we're behind

	pthread_mutex_lock(&eventListenerMapMutex);
	for(unsigned i = 0; i <= std::max(::gamepad_reads_behind, ::remote_reads_behind); ++i) {
		// Which buffer to read from
		::buffer = i;

		typedef std::vector<DispatchFunctor *>::iterator J;
		for(J j = this->dispatch_functor_pointers.begin(); j != this->dispatch_functor_pointers.end(); ++j) {
			(*(*j))();
		}
	}

	input->ClearCachedRemoteReads(0, cached_remote_count0);
	input->ClearCachedRemoteReads(1, cached_remote_count1);
	input->ClearCachedRemoteReads(2, cached_remote_count2);
	input->ClearCachedRemoteReads(3, cached_remote_count3);
	input->ClearCachedGamePadReads(0, cached_gamepad_count0);

	for(unsigned i = 0; i < ::delayed_dispatches->size(); ++i) {
		JSExtensionEvent js_event(WIIU_EMPTY_EVENT, (*delayed_dispatches)[i].event_name, (void *)(*delayed_dispatches)[i].data, 
			(*delayed_dispatches)[i].event_name, (*delayed_dispatches)[i].deviceID, (*delayed_dispatches)[i].controlID, (*delayed_dispatches)[i].indexID); 
		JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(&js_event, (*delayed_dispatches)[i].event_string);
	}

	delayed_dispatches->clear();

	pthread_mutex_unlock(&eventListenerMapMutex);

	if(call_crash_callback) {
		this->crashing = false;
		this->crash_callback();
	}
}

void JSSystemEventDispatcher::DispatchSystemEvent(std::string event_name, std::string event_string, void *data, WiiUEventType_t event_type) {
	if (g_DestructingJSSystemEventDispatcher) return;
	JSEvent js_event(event_type, event_name, (void *)data); 
	this->DispatchEvent(js_event, event_string);
}

void JSSystemEventDispatcher::DispatchExtensionEvent(JSExtensionEvent* jsExtensionEvent, std::string event_string)
{
	if (g_DestructingJSSystemEventDispatcher) return;
	// Retrieve the list of listeners for the given event's type.
	pthread_mutex_lock(&eventListenerMapMutex);
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_string);

	// Send given event to all listeners of its type.
	if(mit != this->event_listener_map.end()) {
		// Creating a copy of the vector here is horribly inefficient. However, it is necessary.
		// It is possible for the JS callback that is called by the below call to OnEvent() to
		// call addEventListener, which adds a listener to event_listener_map (all in the same
		// thread, so mutexes won't work). If event_listener_map gets event listeners added to
		// it while it's being iterated over right now, crash.
		std::vector<JSEventListener *> listeners = *(mit->second);
		JSExtensionEventListener* targetListener;
		for(std::vector<JSEventListener *>::iterator i = listeners.begin(); i != listeners.end(); ++i) {
			targetListener = static_cast<JSExtensionEventListener*>(*i);
			if(targetListener)
				targetListener->OnEvent(jsExtensionEvent);
		}
	}
	pthread_mutex_unlock(&eventListenerMapMutex);
}

void JSSystemEventDispatcher::DispatchExtensionEventThreadSafe(JSExtensionEvent* jsExtensionEvent, std::string event_string)
{
	if (g_DestructingJSSystemEventDispatcher) return;
	
	//Check to see if the event listener is queued added in the same frame as dispatch
	JSExtensionEventListener* targetListener;
	std::vector<ListenerEventPair> listenerAddListCopy;
	typedef std::vector<ListenerEventPair>::iterator K;
	pthread_mutex_lock(&eventListenerMapMutex);
	pthread_mutex_lock(&addListMutex);
	listenerAddListCopy.insert(listenerAddListCopy.begin(), this->listener_add_list.begin(), this->listener_add_list.end());
	pthread_mutex_unlock(&addListMutex);

	for(K k = listenerAddListCopy.begin(); k != listenerAddListCopy.end(); ++k) {

		if(k->event_type == event_string)
		{
			targetListener = static_cast<JSExtensionEventListener*>(k->js_listener);
			if(targetListener)
				targetListener->OnEvent(jsExtensionEvent);
		}
	}

	std::vector<ListenerEventFunctorTuple> addListCopy;
	typedef std::vector<ListenerEventFunctorTuple>::iterator M;
	pthread_mutex_lock(&addListMutex);
	addListCopy.insert(addListCopy.begin(), this->add_list.begin(), this->add_list.end());
	pthread_mutex_unlock(&addListMutex);

	for(M m = addListCopy.begin(); m != addListCopy.end(); ++m) {
		
		if(m->event_type == event_string)
		{
			targetListener = static_cast<JSExtensionEventListener*>(m->js_listener);
			if(targetListener)
				targetListener->OnEvent(jsExtensionEvent);
		}
	}
	// Retrieve the list of listeners for the given event's type.
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_string);

	// Send given event to all listeners of its type.
	if(mit != this->event_listener_map.end()) {
		// Creating a copy of the vector here is horribly inefficient. However, it is necessary.
		// It is possible for the JS callback that is called by the below call to OnEvent() to
		// call addEventListener, which adds a listener to event_listener_map (all in the same
		// thread, so mutexes won't work). If event_listener_map gets event listeners added to
		// it while it's being iterated over right now, crash.
		std::vector<JSEventListener *> listeners = *(mit->second);
		for(std::vector<JSEventListener *>::iterator i = listeners.begin(); i != listeners.end(); ++i) {
			targetListener = static_cast<JSExtensionEventListener*>(*i);
			if(targetListener)
				targetListener->OnEvent(jsExtensionEvent);
		}
	}
	pthread_mutex_unlock(&eventListenerMapMutex);
}

void JSSystemEventDispatcher::DispatchExtensionEvent(std::string event_name, std::string event_string, void *data, 
	WiiUSystemChannel indexID, WiiUSystemDevice deviceID, WiiUSystemControl controlID)
{
	if (g_DestructingJSSystemEventDispatcher) return;
	JSExtensionEvent jsExtensionEvent(WIIU_EMPTY_EVENT, event_name, data, event_string, deviceID, controlID, indexID);
	this->DispatchExtensionEvent(&jsExtensionEvent, event_string);
}


void JSSystemEventDispatcher::QueueEventForDispatch(
	std::string event_name, 
	WiiUSystemDevice device,
	WiiUSystemChannel channel,
	WiiUSystemControl control,
	JSEvent *event
) {
	if (g_DestructingJSSystemEventDispatcher) return;
	QueueDispatchFunctorEvent(
		event_name.c_str(),
		JSSystemEventDispatcher::GetInstance()->GetEventString(
			device, 
			channel,
			control,
			event_name.c_str()
		), 
		event,
		channel,
		device,
		control
	);
}

void JSSystemEventDispatcher::DispatchDisplayEvent(std::string event_string, WiiUSystemChannel channel) {
	if (g_DestructingJSSystemEventDispatcher) return;
	QueueDispatchFunctorEvent(
		event_string.c_str(),
		JSSystemEventDispatcher::GetInstance()->GetEventString(
			DISPLAY, 
			channel,
			DEFAULT,
			event_string.c_str()
		), 
		NULL,
		channel,
		DISPLAY,
		DEFAULT
	);
}

void JSSystemEventDispatcher::DispatchErrorEvent(JSEvent* js_event) {
	if (g_DestructingJSSystemEventDispatcher) return;
	QueueDispatchFunctorEvent(
		"errorWindowClose",
		JSSystemEventDispatcher::GetInstance()->GetEventString(
			ERROR_WINDOW, 
			CHANNEL_0,
			DEFAULT,
			"errorWindowClose"
		), 
		js_event->GetData(),
		CHANNEL_0,
		ERROR_WINDOW,
		DEFAULT
	);
}

void JSSystemEventDispatcher::DispatchCrashEvent(void) {
	if (g_DestructingJSSystemEventDispatcher) return;
	JSSystemEventDispatcher::GetInstance()->DispatchExtensionEvent(
		"crash",
		JSSystemEventDispatcher::GetInstance()->GetEventString(
			SYSTEM, 
			"crash"
		), 
		0,
		CHANNEL_0,
		SYSTEM
	);
}

void JSSystemEventDispatcher::SetCrash(bool crash) {
	this->crashing = crash;
}

bool JSSystemEventDispatcher::GetCrash(void) {
	return this->crashing;
}

void JSSystemEventDispatcher::SetCrashCallback(void (*callback)(void)) {
	this->crash_callback = callback;
}
