/*

	Paired Single Implementation of FFT Code

*/

#include <stdlib.h>
#include <string.h>
#include "libavutil/mathematics.h"
#include "../fft.h"
#include "../fft-internal.h"

#include <ppc.h>
#include <ppc_ps.h>
#include <wkaudio/wkaudAudio.h>

// ----------------------------------------------------------------------------------------------
/**
	Permute data (in preparation to FFT pass) in place
*/
void ff_fft_permute_ps(FFTContext *s, FFTComplex *z, FFTComplex *tmp, int bInPlace)
{
    int j, np;
    const uint16_t *revtab = s->revtab;
    np = 1 << s->nbits;

	if (np < 4)
	{
		for(j=np-1; j>=0; --j)
			__PSQ_ST(&tmp[revtab[j]], __PSQ_L(&z[j], 0, 0), 0, 0);
	
		if (bInPlace)
		{
			for(j=np-1; j>=0; --j)
				__PSQ_ST(&z[j], __PSQ_L(&tmp[j], 0, 0), 0, 0);
		}
	}
	else
	{
		np >>= 2;

		const f32x2 *pSrc = (const f32x2*)z;
		const uint16_t *pRev = revtab;
		for(j=np-1; j>=0; --j)
		{
			int di0 = pRev[0];
			int di1 = pRev[1];
			int di2 = pRev[2];
			int di3 = pRev[3];
			pRev += 4;
			f32x2 in0 = __PSQ_L(&pSrc[0], 0, 0);
			f32x2 in1 = __PSQ_L(&pSrc[1], 0, 0);
			f32x2 in2 = __PSQ_L(&pSrc[2], 0, 0);
			f32x2 in3 = __PSQ_L(&pSrc[3], 0, 0);
			pSrc += 4;
			__PSQ_ST(&tmp[di0], in0, 0, 0);
			__PSQ_ST(&tmp[di1], in1, 0, 0);
			__PSQ_ST(&tmp[di2], in2, 0, 0);
			__PSQ_ST(&tmp[di3], in3, 0, 0);
		}

		if (bInPlace)
		{
			pSrc = (const f32x2*)tmp;
			f32x2 *pDst = (f32x2*)z;
			for(j=np-1; j>=0; --j)
			{
				f32x2 in0 = __PSQ_L(&pSrc[0], 0, 0);
				f32x2 in1 = __PSQ_L(&pSrc[1], 0, 0);
				f32x2 in2 = __PSQ_L(&pSrc[2], 0, 0);
				f32x2 in3 = __PSQ_L(&pSrc[3], 0, 0);
				pSrc += 4;
				__PSQ_ST(&pDst[0], in0, 0, 0);
				__PSQ_ST(&pDst[1], in1, 0, 0);
				__PSQ_ST(&pDst[2], in2, 0, 0);
				__PSQ_ST(&pDst[3], in3, 0, 0);
				pDst += 4;
			}
		}
	}
}

// Version of permutate that assume the 2nd half of the input buffer as zero!
// (it also will keep the data inside the tmp buffer at all times!)
void ff_fft_permute_2ndzero_ps(FFTContext *s, FFTComplex *z, FFTComplex *tmp)
{
    int j, np;
    const uint16_t *revtab = s->revtab;
    np = 1 << s->nbits;

	if (np < 4)
	{
		// for small data we still just operate normally...
		for(j=np-1; j>=0; --j)
			__PSQ_ST(&tmp[revtab[j]], __PSQ_L(&z[j], 0, 0), 0, 0);
	}
	else
	{
		// Clear out all of the temp buffer, so we can partially update it
		FFTComplex *pDst = tmp;
		for(j=(np*sizeof(FFTComplex)/32)-1; j>=0; --j)
		{
			__DCBZ(pDst,0);
			pDst += 32/sizeof(FFTComplex);
		}

		np >>= (2 + 1);		// only work on the 1st half of the data!

		const f32x2 *pSrc = (const f32x2*)z;
		const uint16_t *pRev = revtab;
		for(j=np-1; j>=0; --j)
		{
			int di0 = pRev[0];
			int di1 = pRev[1];
			int di2 = pRev[2];
			int di3 = pRev[3];
			pRev += 4;
			f32x2 in0 = __PSQ_L(&pSrc[0], 0, 0);
			f32x2 in1 = __PSQ_L(&pSrc[1], 0, 0);
			f32x2 in2 = __PSQ_L(&pSrc[2], 0, 0);
			f32x2 in3 = __PSQ_L(&pSrc[3], 0, 0);
			pSrc += 4;
			__PSQ_ST(&tmp[di0], in0, 0, 0);
			__PSQ_ST(&tmp[di1], in1, 0, 0);
			__PSQ_ST(&tmp[di2], in2, 0, 0);
			__PSQ_ST(&tmp[di3], in3, 0, 0);
		}
	}
}

void ff_fft_permute_ps_cb(FFTContext *s, FFTComplex *z)
{
	void *pLC = wksupAudioLCBeginUse((1 << s->nbits) * sizeof(FFTComplex));
	ff_fft_permute_ps(s, z, (FFTComplex*)pLC, 1);
	wksupAudioLCEndUse();
}

// ----------------------------------------------------------------------------------------------

void ff_fft_permute_fromseparate_ps(FFTContext *s, const FFTSample *r, const FFTSample *i, FFTComplex *tmp)
{
    int j, np;
    const uint16_t *revtab = s->revtab;
    np = 1 << s->nbits;

	if (np < 4)
	{
		for(j=np-1; j>=0; --j)
			__PSQ_ST(&tmp[revtab[j]], __PS_MERGE00(__PSQ_L(&r[j], 1, 0), __PSQ_L(&i[j], 1, 0)), 0, 0);
	}
	else
	{
		np >>= 2;

		const FFTSample *pSrcR = r;
		const FFTSample *pSrcI = i;
		const uint16_t *pRev = revtab;
		for(j=np-1; j>=0; --j)
		{
			f32x2 real0 = __PSQ_L(&pSrcR[0], 0, 0);
			f32x2 imag0 = __PSQ_L(&pSrcI[0], 0, 0);
			f32x2 real1 = __PSQ_L(&pSrcR[2], 0, 0);
			f32x2 imag1 = __PSQ_L(&pSrcI[2], 0, 0);
			f32x2 in0 = __PS_MERGE00(real0, imag0);
			int di0 = pRev[0];
			f32x2 in1 = __PS_MERGE11(real0, imag0);
			int di1 = pRev[1];
			f32x2 in2 = __PS_MERGE00(real1, imag1);
			int di2 = pRev[2];
			f32x2 in3 = __PS_MERGE11(real1, imag1);
			int di3 = pRev[3];
			__PSQ_ST(&tmp[di0], in0, 0, 0);
			pSrcR += 4;
			__PSQ_ST(&tmp[di1], in1, 0, 0);
			pSrcI += 4;
			__PSQ_ST(&tmp[di2], in2, 0, 0);
			pRev += 4;
			__PSQ_ST(&tmp[di3], in3, 0, 0);
		}
	}
}

// ----------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------
#if 1
static const f32x2 cnst_m1p1 = {-1.0f,  1.0f};
static const f32x2 cnst_p1m1 = {1.0f,  -1.0f};
static const f32x2 cnst_sqrthalf = {sqrthalf, sqrthalf};

#define PS_CMUL(d, a, b) \
	{ d = __PS_MADD(__PS_MERGE10(b, b), __PS_MULS1(cnst_m1p1, a), __PS_MULS0(b, a)); }

#define PS_BUTTERFLIES(a0, a1, a2, a3, t12, t56) \
{												\
	f32x2 tmp25 = __PS_MERGE10(t12, t56);		\
	f32x2 tmp61 = __PS_MERGE10(t56, t12); 		\
	f32x2 tmp43 = __PS_SUB(tmp25, tmp61);		\
	f32x2 tmp65 = __PS_ADD(tmp25, tmp61);		\
	f32x2 tmp56 = __PS_MERGE10(tmp65, tmp65);	\
	a3 = __PS_SUB(a1, tmp43);					\
	a2 = __PS_SUB(a0, tmp56);					\
	a1 = __PS_ADD(a1, tmp43);					\
	a0 = __PS_ADD(a0, tmp56);					\
}

#define PS_TRANSFORM(a0,a1,a2,a3,w) {\
	f32x2 tmp12, tmp56; \
    PS_CMUL(tmp12, a2, __PS_MUL(w, cnst_p1m1));\
    PS_CMUL(tmp56, a3, w);\
    PS_BUTTERFLIES(a0,a1,a2,a3,tmp12,tmp56)\
}

#define PS_TRANSFORM_ZERO(a0,a1,a2,a3) {\
	f32x2 tmp12 = a2; \
	f32x2 tmp56 = a3; \
    PS_BUTTERFLIES(a0,a1,a2,a3,tmp12,tmp56)\
}

/* z[0...8n-1], w[1...2n-1] */
#define PS_PASS(name)\
static inline void name(FFTComplex *z, const FFTSample *wre, unsigned int n)\
{\
    int o1 = 2*n;\
    int o2 = 4*n;\
    int o3 = 6*n;\
    const FFTSample *wim = wre+o1;\
    n--;\
\
	f32x2 ai0 = __PSQ_L(&z[0], 0, 0); \
	f32x2 ai1 = __PSQ_L(&z[o1], 0, 0); \
	f32x2 ai2 = __PSQ_L(&z[o2], 0, 0); \
	f32x2 ai3 = __PSQ_L(&z[o3], 0, 0); \
    PS_TRANSFORM_ZERO(ai0, ai1, ai2, ai3);\
	__PSQ_ST(&z[0], ai0, 0, 0); \
	__PSQ_ST(&z[o1], ai1, 0, 0); \
	__PSQ_ST(&z[o2], ai2, 0, 0); \
	__PSQ_ST(&z[o3], ai3, 0, 0); \
	f32x2 wi; \
	wi[0] = wre[1]; \
	wi[1] = wim[-1]; \
	ai0 = __PSQ_L(&z[1], 0, 0); \
	ai1 = __PSQ_L(&z[o1+1], 0, 0); \
	ai2 = __PSQ_L(&z[o2+1], 0, 0); \
	ai3 = __PSQ_L(&z[o3+1], 0, 0); \
    PS_TRANSFORM(ai0, ai1, ai2, ai3, wi);\
	__PSQ_ST(&z[1], ai0, 0, 0); \
	__PSQ_ST(&z[o1+1], ai1, 0, 0); \
	__PSQ_ST(&z[o2+1], ai2, 0, 0); \
	__PSQ_ST(&z[o3+1], ai3, 0, 0); \
    do {\
        z += 2;\
        wre += 2;\
        wim -= 2;\
		f32x2 vwre = __PSQ_L(&wre[0], 0, 0); \
		f32x2 vwim = __PSQ_L(&wim[-1], 0, 0); \
		ai0 = __PSQ_L(&z[0], 0, 0); \
		ai1 = __PSQ_L(&z[o1], 0, 0); \
		ai2 = __PSQ_L(&z[o2], 0, 0); \
		ai3 = __PSQ_L(&z[o3], 0, 0); \
        PS_TRANSFORM(ai0, ai1, ai2, ai3, __PS_MERGE01(vwre, vwim)); \
		__PSQ_ST(&z[0], ai0, 0, 0); \
		__PSQ_ST(&z[o1], ai1, 0, 0); \
		__PSQ_ST(&z[o2], ai2, 0, 0); \
		__PSQ_ST(&z[o3], ai3, 0, 0); \
		ai0 = __PSQ_L(&z[1], 0, 0); \
		ai1 = __PSQ_L(&z[o1+1], 0, 0); \
		ai2 = __PSQ_L(&z[o2+1], 0, 0); \
		ai3 = __PSQ_L(&z[o3+1], 0, 0); \
        PS_TRANSFORM(ai0, ai1, ai2, ai3, __PS_MERGE10(vwre, vwim)); \
		__PSQ_ST(&z[1], ai0, 0, 0); \
		__PSQ_ST(&z[o1+1], ai1, 0, 0); \
		__PSQ_ST(&z[o2+1], ai2, 0, 0); \
		__PSQ_ST(&z[o3+1], ai3, 0, 0); \
    } while(--n);\
}

PS_PASS(pass)

#define PS_FFT4(z0, z1, z2, z3) \
{ \
	f32x2 t12 = __PS_ADD(z0, z1); \
	f32x2 t34 = __PS_SUB(z0, z1); \
\
	f32x2 tmp0 = __PS_MERGE10(z2, z3); \
	f32x2 tmp1 = __PS_MERGE10(z3, z2); \
\
	f32x2 t56 = __PS_ADD(tmp0, tmp1); \
	f32x2 t78 = __PS_SUB(tmp0, tmp1); \
\
	f32x2 t65 = __PS_MERGE10(t56, t56); \
\
	z0 = __PS_ADD(t12, t65); \
	z1 = __PS_ADD(t34, t78); \
	z2 = __PS_SUB(t12, t65); \
	z3 = __PS_SUB(t34, t78); \
}

static inline void ps_fft4(FFTComplex *z)
{
	f32x2 z0 = __PSQ_L(&z[0], 0, 0);
	f32x2 z1 = __PSQ_L(&z[1], 0, 0);
	f32x2 z2 = __PSQ_L(&z[2], 0, 0);
	f32x2 z3 = __PSQ_L(&z[3], 0, 0);

	PS_FFT4(z0, z1, z2, z3);

	__PSQ_ST(&z[0], z0, 0, 0);
	__PSQ_ST(&z[1], z1, 0, 0);
	__PSQ_ST(&z[2], z2, 0, 0);
	__PSQ_ST(&z[3], z3, 0, 0);
}

#define PS_FFT8(ai0, ai1, ai2, ai3, ai4, ai5, ai6, ai7) \
{ \
    PS_FFT4(ai0, ai1, ai2, ai3); \
 \
	f32x2 t12 = __PS_ADD(ai4, ai5); \
	f32x2 t56 = __PS_ADD(ai6, ai7); \
 \
	ai5 = __PS_SUB(ai4, ai5); \
	ai7 = __PS_SUB(ai6, ai7); \
 \
    PS_BUTTERFLIES(ai0, ai2, ai4, ai6, t12, t56); \
 \
    PS_TRANSFORM(ai1, ai3, ai5, ai7, cnst_sqrthalf); \
}

static inline void ps_fft8(FFTComplex *z)
{
	f32x2 ai0 = __PSQ_L(&z[0], 0, 0);
	f32x2 ai1 = __PSQ_L(&z[1], 0, 0);
	f32x2 ai2 = __PSQ_L(&z[2], 0, 0);
	f32x2 ai3 = __PSQ_L(&z[3], 0, 0);
	f32x2 ai4 = __PSQ_L(&z[4], 0, 0);
	f32x2 ai5 = __PSQ_L(&z[5], 0, 0);
	f32x2 ai6 = __PSQ_L(&z[6], 0, 0);
	f32x2 ai7 = __PSQ_L(&z[7], 0, 0);

	PS_FFT8(ai0, ai1, ai2, ai3, ai4, ai5, ai6, ai7);

	__PSQ_ST(&z[0], ai0, 0, 0);
	__PSQ_ST(&z[1], ai1, 0, 0);
	__PSQ_ST(&z[2], ai2, 0, 0);
	__PSQ_ST(&z[3], ai3, 0, 0);
	__PSQ_ST(&z[4], ai4, 0, 0);
	__PSQ_ST(&z[5], ai5, 0, 0);
	__PSQ_ST(&z[6], ai6, 0, 0);
	__PSQ_ST(&z[7], ai7, 0, 0);
}

static inline void ps_fft16(FFTComplex *z)
{
    FFTDouble t1, t2, t3, t4, t5, t6;
    FFTSample cos_16_1 = FFT_NAME(ff_cos_16)[1];
    FFTSample cos_16_3 = FFT_NAME(ff_cos_16)[3];

	f32x2 ai0 = __PSQ_L(&z[0], 0, 0);
	f32x2 ai1 = __PSQ_L(&z[1], 0, 0);
	f32x2 ai2 = __PSQ_L(&z[2], 0, 0);
	f32x2 ai3 = __PSQ_L(&z[3], 0, 0);
	f32x2 ai4 = __PSQ_L(&z[4], 0, 0);
	f32x2 ai5 = __PSQ_L(&z[5], 0, 0);
	f32x2 ai6 = __PSQ_L(&z[6], 0, 0);
	f32x2 ai7 = __PSQ_L(&z[7], 0, 0);
	f32x2 ai8 = __PSQ_L(&z[8], 0, 0);
	f32x2 ai9 = __PSQ_L(&z[9], 0, 0);
	f32x2 ai10 = __PSQ_L(&z[10], 0, 0);
	f32x2 ai11 = __PSQ_L(&z[11], 0, 0);
	f32x2 ai12 = __PSQ_L(&z[12], 0, 0);
	f32x2 ai13 = __PSQ_L(&z[13], 0, 0);
	f32x2 ai14 = __PSQ_L(&z[14], 0, 0);
	f32x2 ai15 = __PSQ_L(&z[15], 0, 0);

	PS_FFT8(ai0, ai1, ai2, ai3, ai4, ai5, ai6, ai7);
	PS_FFT4(ai8, ai9, ai10, ai11);
	PS_FFT4(ai12, ai13, ai14, ai15);

    PS_TRANSFORM_ZERO(ai0, ai4, ai8, ai12);
    PS_TRANSFORM(ai2, ai6, ai10, ai14, cnst_sqrthalf);

	f32x2 wi;
	wi[0] = cos_16_1;
	wi[1] = cos_16_3;
    PS_TRANSFORM(ai1, ai5, ai9, ai13, wi);
    PS_TRANSFORM(ai3, ai7, ai11, ai15, __PS_MERGE10(wi, wi));

	__PSQ_ST(&z[0], ai0, 0, 0);
	__PSQ_ST(&z[1], ai1, 0, 0);
	__PSQ_ST(&z[2], ai2, 0, 0);
	__PSQ_ST(&z[3], ai3, 0, 0);
	__PSQ_ST(&z[4], ai4, 0, 0);
	__PSQ_ST(&z[5], ai5, 0, 0);
	__PSQ_ST(&z[6], ai6, 0, 0);
	__PSQ_ST(&z[7], ai7, 0, 0);
	__PSQ_ST(&z[8], ai8, 0, 0);
	__PSQ_ST(&z[9], ai9, 0, 0);
	__PSQ_ST(&z[10], ai10, 0, 0);
	__PSQ_ST(&z[11], ai11, 0, 0);
	__PSQ_ST(&z[12], ai12, 0, 0);
	__PSQ_ST(&z[13], ai13, 0, 0);
	__PSQ_ST(&z[14], ai14, 0, 0);
	__PSQ_ST(&z[15], ai15, 0, 0);
}


#define DECL_FFT(n,n2,n4)\
static inline void ps_fft##n(FFTComplex *z)\
{\
    ps_fft##n2(z);\
    ps_fft##n4(z+n4*2);\
    ps_fft##n4(z+n4*3);\
    pass(z,FFT_NAME(ff_cos_##n),n4/2);\
}

DECL_FFT(32,16,8)
DECL_FFT(64,32,16)
DECL_FFT(128,64,32)
DECL_FFT(256,128,64)
DECL_FFT(512,256,128)
DECL_FFT(1024,512,256)
DECL_FFT(2048,1024,512)
DECL_FFT(4096,2048,1024)
DECL_FFT(8192,4096,2048)
DECL_FFT(16384,8192,4096)
DECL_FFT(32768,16384,8192)
DECL_FFT(65536,32768,16384)

static void (* const ps_fft_dispatch[])(FFTComplex*) = {
    ps_fft4, ps_fft8, ps_fft16, ps_fft32, ps_fft64, ps_fft128, ps_fft256, ps_fft512, ps_fft1024,
    ps_fft2048, ps_fft4096, ps_fft8192, ps_fft16384, ps_fft32768, ps_fft65536,
};

void ff_fft_calc_ps(FFTContext *s, FFTComplex *z)
{
    ps_fft_dispatch[s->nbits-2](z);
}

#else

#define BUTTERFLIES(a0,a1,a2,a3) {\
    BF(t3, t5, t5, t1);\
    BF(a2.re, a0.re, a0.re, t5);\
    BF(a3.im, a1.im, a1.im, t3);\
    BF(t4, t6, t2, t6);\
    BF(a3.re, a1.re, a1.re, t4);\
    BF(a2.im, a0.im, a0.im, t6);\
}

// force loading all the inputs before storing any.
// this is slightly slower for small data, but avoids store->load aliasing
// for addresses separated by large powers of 2.
#define BUTTERFLIES_BIG(a0,a1,a2,a3) {\
    FFTSample r0=a0.re, i0=a0.im, r1=a1.re, i1=a1.im;\
    BF(t3, t5, t5, t1);\
    BF(a2.re, a0.re, r0, t5);\
    BF(a3.im, a1.im, i1, t3);\
    BF(t4, t6, t2, t6);\
    BF(a3.re, a1.re, r1, t4);\
    BF(a2.im, a0.im, i0, t6);\
}

#define TRANSFORM(a0,a1,a2,a3,wre,wim) {\
    CMUL(t1, t2, a2.re, a2.im, wre, -wim);\
    CMUL(t5, t6, a3.re, a3.im, wre,  wim);\
    BUTTERFLIES(a0,a1,a2,a3)\
}

#define TRANSFORM_ZERO(a0,a1,a2,a3) {\
    t1 = a2.re;\
    t2 = a2.im;\
    t5 = a3.re;\
    t6 = a3.im;\
    BUTTERFLIES(a0,a1,a2,a3)\
}

/* z[0...8n-1], w[1...2n-1] */
#define PASS(name)\
static void name(FFTComplex *z, const FFTSample *wre, unsigned int n)\
{\
    FFTDouble t1, t2, t3, t4, t5, t6;\
    int o1 = 2*n;\
    int o2 = 4*n;\
    int o3 = 6*n;\
    const FFTSample *wim = wre+o1;\
    n--;\
\
    TRANSFORM_ZERO(z[0],z[o1],z[o2],z[o3]);\
    TRANSFORM(z[1],z[o1+1],z[o2+1],z[o3+1],wre[1],wim[-1]);\
    do {\
        z += 2;\
        wre += 2;\
        wim -= 2;\
        TRANSFORM(z[0],z[o1],z[o2],z[o3],wre[0],wim[0]);\
        TRANSFORM(z[1],z[o1+1],z[o2+1],z[o3+1],wre[1],wim[-1]);\
    } while(--n);\
}

PASS(pass)
#undef BUTTERFLIES
#define BUTTERFLIES BUTTERFLIES_BIG
PASS(pass_big)

#define DECL_FFT(n,n2,n4)\
static void fft##n(FFTComplex *z)\
{\
    fft##n2(z);\
    fft##n4(z+n4*2);\
    fft##n4(z+n4*3);\
    pass(z,FFT_NAME(ff_cos_##n),n4/2);\
}

static void fft4(FFTComplex *z)
{
    FFTDouble t1, t2, t3, t4, t5, t6, t7, t8;

    BF(t3, t1, z[0].re, z[1].re);
    BF(t8, t6, z[3].re, z[2].re);
    BF(z[2].re, z[0].re, t1, t6);
    BF(t4, t2, z[0].im, z[1].im);
    BF(t7, t5, z[2].im, z[3].im);
    BF(z[3].im, z[1].im, t4, t8);
    BF(z[3].re, z[1].re, t3, t7);
    BF(z[2].im, z[0].im, t2, t5);
}

static void fft8(FFTComplex *z)
{
    FFTDouble t1, t2, t3, t4, t5, t6;

    fft4(z);

    BF(t1, z[5].re, z[4].re, -z[5].re);
    BF(t2, z[5].im, z[4].im, -z[5].im);
    BF(t5, z[7].re, z[6].re, -z[7].re);
    BF(t6, z[7].im, z[6].im, -z[7].im);

    BUTTERFLIES(z[0],z[2],z[4],z[6]);
    TRANSFORM(z[1],z[3],z[5],z[7],sqrthalf,sqrthalf);
}

#if !CONFIG_SMALL
static void fft16(FFTComplex *z)
{
    FFTDouble t1, t2, t3, t4, t5, t6;
    FFTSample cos_16_1 = FFT_NAME(ff_cos_16)[1];
    FFTSample cos_16_3 = FFT_NAME(ff_cos_16)[3];

    fft8(z);
    fft4(z+8);
    fft4(z+12);

    TRANSFORM_ZERO(z[0],z[4],z[8],z[12]);
    TRANSFORM(z[2],z[6],z[10],z[14],sqrthalf,sqrthalf);
    TRANSFORM(z[1],z[5],z[9],z[13],cos_16_1,cos_16_3);
    TRANSFORM(z[3],z[7],z[11],z[15],cos_16_3,cos_16_1);
}
#else
DECL_FFT(16,8,4)
#endif
DECL_FFT(32,16,8)
DECL_FFT(64,32,16)
DECL_FFT(128,64,32)
DECL_FFT(256,128,64)
DECL_FFT(512,256,128)
#if !CONFIG_SMALL
#define pass pass_big
#endif
DECL_FFT(1024,512,256)
DECL_FFT(2048,1024,512)
DECL_FFT(4096,2048,1024)
DECL_FFT(8192,4096,2048)
DECL_FFT(16384,8192,4096)
DECL_FFT(32768,16384,8192)
DECL_FFT(65536,32768,16384)

static void (* const fft_dispatch[])(FFTComplex*) = {
    fft4, fft8, fft16, fft32, fft64, fft128, fft256, fft512, fft1024,
    fft2048, fft4096, fft8192, fft16384, fft32768, fft65536,
};

void ff_fft_calc_ps(FFTContext *s, FFTComplex *z)
{
    fft_dispatch[s->nbits-2](z);
}

#endif

// -----------------------------------------------------------------------------------------------

void ff_fft_init_ps(FFTContext *s)
{
    s->fft_permute = ff_fft_permute_ps_cb;
	s->fft_calc = ff_fft_calc_ps;
}


